"""Tests for coderace race mode."""

from __future__ import annotations

import json
import time
from pathlib import Path
from unittest.mock import patch

from typer.testing import CliRunner

from coderace.cli import app
from coderace.commands.race import (
    RaceParticipant,
    RaceSummary,
    _build_live_panel,
    _resolve_verify_commands,
    _save_race_summary,
    _summary_record,
    _winner_announcement,
    run_race,
)
from coderace.types import AgentResult, Task

runner = CliRunner()


def _make_task(tmp_repo: Path, agents: list[str]) -> Task:
    return Task(
        name="race-task",
        description="Fix race task",
        repo=tmp_repo,
        test_command="echo ok",
        agents=agents,
        timeout=60,
    )


def test_run_race_first_verify_pass_wins(tmp_repo: Path) -> None:
    task = _make_task(tmp_repo, ["claude", "codex"])

    plans = {
        "claude": {"duration": 0.04, "verify_passed": True},
        "codex": {"duration": 0.20, "verify_passed": True},
    }

    def fake_runner(**kwargs):
        agent = kwargs["agent_name"]
        stop_event = kwargs["stop_event"]
        status_callback = kwargs["status_callback"]
        status_callback("coding")
        deadline = time.monotonic() + plans[agent]["duration"]
        while time.monotonic() < deadline:
            if stop_event.is_set():
                return (
                    AgentResult(
                        agent=agent,
                        exit_code=130,
                        stdout="",
                        stderr="stopped",
                        wall_time=0.01,
                        timed_out=False,
                    ),
                    0,
                    {"stopped": True, "total_wall_time": 0.01},
                )
            time.sleep(0.005)
        status_callback("testing")
        return (
            AgentResult(
                agent=agent,
                exit_code=0,
                stdout="",
                stderr="",
                wall_time=plans[agent]["duration"],
                timed_out=False,
            ),
            10,
            {
                "verify_passed": plans[agent]["verify_passed"],
                "verify_exit_codes": [0],
                "verify_outputs": ["ok"],
                "stopped": False,
                "total_wall_time": plans[agent]["duration"] + 0.01,
            },
        )

    with patch("coderace.commands.race._invoke_worktree_runner", side_effect=fake_runner):
        summary = run_race(
            task=task,
            base_ref="abc123",
            timeout=60,
            no_cost=True,
            verify_commands=["pytest verify.py -q"],
            poll_interval_seconds=0.01,
            refresh_interval_seconds=0.01,
            graceful_shutdown_seconds=0.10,
        )

    assert summary.winner_agent == "claude"
    winner = next(p for p in summary.participants if p.agent == "claude")
    assert winner.status == "winner"
    assert winner.verify_passed is True
    loser = next(p for p in summary.participants if p.agent == "codex")
    assert loser.status in {"stopped", "winner"}


def test_run_race_no_winner_when_all_agents_fail(tmp_repo: Path) -> None:
    task = _make_task(tmp_repo, ["claude", "codex"])

    def fake_runner(**kwargs):
        agent = kwargs["agent_name"]
        return (
            AgentResult(
                agent=agent,
                exit_code=1,
                stdout="",
                stderr="failed",
                wall_time=0.01,
                timed_out=False,
            ),
            0,
            {
                "verify_passed": False,
                "verify_exit_codes": [1],
                "verify_outputs": ["failed"],
                "stopped": False,
                "total_wall_time": 0.01,
            },
        )

    with patch("coderace.commands.race._invoke_worktree_runner", side_effect=fake_runner):
        summary = run_race(
            task=task,
            base_ref="abc123",
            timeout=60,
            no_cost=True,
            verify_commands=["pytest verify.py -q"],
            poll_interval_seconds=0.01,
            refresh_interval_seconds=0.01,
            graceful_shutdown_seconds=0.05,
        )

    assert summary.winner_agent is None
    assert all(p.status in {"failed", "failed_verify"} for p in summary.participants)


def test_run_race_cancels_remaining_agents_after_winner(tmp_repo: Path) -> None:
    task = _make_task(tmp_repo, ["claude", "codex", "aider"])

    plans = {
        "claude": {"duration": 0.03, "exit_code": 0},
        "codex": {"duration": 0.30, "exit_code": 0},
        "aider": {"duration": 0.30, "exit_code": 0},
    }

    def fake_runner(**kwargs):
        agent = kwargs["agent_name"]
        stop_event = kwargs["stop_event"]
        deadline = time.monotonic() + plans[agent]["duration"]
        while time.monotonic() < deadline:
            if stop_event.is_set():
                return (
                    AgentResult(
                        agent=agent,
                        exit_code=130,
                        stdout="",
                        stderr="stopped",
                        wall_time=0.02,
                        timed_out=False,
                    ),
                    0,
                    {"stopped": True, "total_wall_time": 0.02},
                )
            time.sleep(0.005)

        return (
            AgentResult(
                agent=agent,
                exit_code=plans[agent]["exit_code"],
                stdout="",
                stderr="",
                wall_time=plans[agent]["duration"],
                timed_out=False,
            ),
            3,
            {
                "verify_passed": True,
                "verify_exit_codes": [0],
                "verify_outputs": ["ok"],
                "stopped": False,
                "total_wall_time": plans[agent]["duration"],
            },
        )

    with patch("coderace.commands.race._invoke_worktree_runner", side_effect=fake_runner):
        summary = run_race(
            task=task,
            base_ref="abc123",
            timeout=60,
            no_cost=True,
            verify_commands=[],
            poll_interval_seconds=0.01,
            refresh_interval_seconds=0.01,
            graceful_shutdown_seconds=0.10,
        )

    assert summary.winner_agent == "claude"
    statuses = {p.agent: p.status for p in summary.participants}
    assert statuses["claude"] == "winner"
    assert "stopped" in {statuses["codex"], statuses["aider"]}


def test_run_race_without_verify_uses_first_clean_exit(tmp_repo: Path) -> None:
    task = _make_task(tmp_repo, ["claude", "codex"])

    plans = {"claude": 0.35, "codex": 0.02}

    def fake_runner(**kwargs):
        agent = kwargs["agent_name"]
        time.sleep(plans[agent])
        return (
            AgentResult(
                agent=agent,
                exit_code=0,
                stdout="",
                stderr="",
                wall_time=plans[agent],
                timed_out=False,
            ),
            1,
            {"stopped": False, "total_wall_time": plans[agent]},
        )

    with patch("coderace.commands.race._invoke_worktree_runner", side_effect=fake_runner):
        summary = run_race(
            task=task,
            base_ref="abc123",
            timeout=60,
            no_cost=True,
            verify_commands=[],
            poll_interval_seconds=0.01,
            refresh_interval_seconds=0.01,
            graceful_shutdown_seconds=0.40,
        )

    assert summary.winner_agent == "codex"


def test_run_race_all_timeouts_produces_no_winner(tmp_repo: Path) -> None:
    task = _make_task(tmp_repo, ["claude", "codex"])

    def fake_runner(**kwargs):
        agent = kwargs["agent_name"]
        return (
            AgentResult(
                agent=agent,
                exit_code=-1,
                stdout="",
                stderr="timed out",
                wall_time=0.02,
                timed_out=True,
            ),
            0,
            {"stopped": False, "total_wall_time": 0.02},
        )

    with patch("coderace.commands.race._invoke_worktree_runner", side_effect=fake_runner):
        summary = run_race(
            task=task,
            base_ref="abc123",
            timeout=60,
            no_cost=True,
            verify_commands=[],
            poll_interval_seconds=0.01,
            refresh_interval_seconds=0.01,
            graceful_shutdown_seconds=0.05,
        )

    assert summary.winner_agent is None
    assert all(p.status == "timed_out" for p in summary.participants)


def test_race_command_is_registered() -> None:
    result = runner.invoke(app, ["race", "--help"])
    assert result.exit_code == 0
    assert "--builtin" in result.output
    assert "--agent" in result.output


def test_race_cli_invocation_with_mock_runner(tmp_repo: Path, tmp_path: Path) -> None:
    task_yaml = tmp_path / "race-task.yaml"
    task_yaml.write_text(
        f"""name: race-task
description: test race
repo: {tmp_repo}
test_command: echo ok
agents:
  - claude
  - codex
""",
        encoding="utf-8",
    )

    summary = RaceSummary(
        race_id="rid",
        task_name="race-task",
        winner_agent="claude",
        winner_time=12.3,
        participants=[],
        timestamp="2026-03-03T00:00:00+00:00",
    )

    with patch("coderace.commands.race.has_uncommitted_changes", return_value=False), patch(
        "coderace.commands.race.get_current_ref", return_value="abc12345"
    ), patch("coderace.commands.race.run_race", return_value=summary), patch(
        "coderace.commands.race.prune_worktrees"
    ):
        result = runner.invoke(
            app,
            ["race", str(task_yaml), "--agent", "claude", "--agent", "codex", "--no-save"],
        )

    assert result.exit_code == 0
    assert "Winner:" in result.output
    assert "claude" in result.output


def test_run_race_emits_status_transitions_to_callback(tmp_repo: Path) -> None:
    task = _make_task(tmp_repo, ["claude"])
    snapshots: list[list[str]] = []

    def fake_runner(**kwargs):
        status_callback = kwargs["status_callback"]
        status_callback("coding")
        time.sleep(0.01)
        status_callback("testing")
        time.sleep(0.01)
        return (
            AgentResult(
                agent="claude",
                exit_code=0,
                stdout="",
                stderr="",
                wall_time=0.02,
                timed_out=False,
            ),
            1,
            {
                "verify_passed": True,
                "verify_exit_codes": [0],
                "verify_outputs": ["ok"],
                "stopped": False,
                "total_wall_time": 0.03,
            },
        )

    def update_callback(participants: list[RaceParticipant]) -> None:
        snapshots.append([p.status for p in participants])

    with patch("coderace.commands.race._invoke_worktree_runner", side_effect=fake_runner):
        summary = run_race(
            task=task,
            base_ref="abc123",
            timeout=60,
            no_cost=True,
            verify_commands=["pytest verify.py -q"],
            update_callback=update_callback,
            poll_interval_seconds=0.01,
            refresh_interval_seconds=0.01,
            graceful_shutdown_seconds=0.10,
        )

    assert summary.winner_agent == "claude"
    flattened = [status for snap in snapshots for status in snap]
    assert "coding" in flattened
    assert "testing" in flattened
    assert "winner" in flattened


def test_race_cli_updates_live_panel_with_mocked_live(tmp_repo: Path, tmp_path: Path) -> None:
    task_yaml = tmp_path / "race-live.yaml"
    task_yaml.write_text(
        f"""name: race-live
description: test live
repo: {tmp_repo}
test_command: echo ok
agents:
  - claude
  - codex
""",
        encoding="utf-8",
    )

    summary = RaceSummary(
        race_id="rid",
        task_name="race-live",
        winner_agent="claude",
        winner_time=0.5,
        participants=[
            RaceParticipant(
                agent="claude",
                status="winner",
                started_at=0.0,
                finished_at=0.5,
                total_time=0.5,
                result=AgentResult(
                    agent="claude",
                    exit_code=0,
                    stdout="",
                    stderr="",
                    wall_time=0.5,
                    timed_out=False,
                ),
            ),
            RaceParticipant(
                agent="codex",
                status="stopped",
                started_at=0.0,
                finished_at=0.8,
                total_time=0.8,
                result=AgentResult(
                    agent="codex",
                    exit_code=130,
                    stdout="",
                    stderr="stopped",
                    wall_time=0.8,
                    timed_out=False,
                ),
            ),
        ],
        timestamp="2026-03-03T00:00:00+00:00",
    )

    live_updates: list[object] = []

    class FakeLive:
        def __init__(self, *args, **kwargs):
            self.initial = args[0] if args else None

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc, tb):
            return False

        def update(self, renderable):
            live_updates.append(renderable)

    def fake_run_race(**kwargs):
        update = kwargs["update_callback"]
        update(
            [
                RaceParticipant(agent="claude", status="coding", started_at=0.0),
                RaceParticipant(agent="codex", status="testing", started_at=0.0),
            ]
        )
        update(
            [
                RaceParticipant(agent="claude", status="winner", started_at=0.0, total_time=0.5),
                RaceParticipant(agent="codex", status="stopped", started_at=0.0, total_time=0.8),
            ]
        )
        return summary

    with patch("coderace.commands.race.has_uncommitted_changes", return_value=False), patch(
        "coderace.commands.race.get_current_ref", return_value="abc12345"
    ), patch("coderace.commands.race.prune_worktrees"), patch(
        "coderace.commands.race.Live", FakeLive
    ), patch("coderace.commands.race.run_race", side_effect=fake_run_race):
        result = runner.invoke(app, ["race", str(task_yaml), "--no-save"])

    assert result.exit_code == 0
    assert "Winner:" in result.output
    assert "Runner-up:" in result.output
    assert len(live_updates) >= 2


def test_summary_record_contains_winner_and_participant_fields() -> None:
    summary = RaceSummary(
        race_id="race-1",
        task_name="task-a",
        winner_agent="claude",
        winner_time=1.23,
        participants=[
            RaceParticipant(
                agent="claude",
                status="winner",
                total_time=1.23,
                lines_changed=7,
                result=AgentResult(
                    agent="claude",
                    exit_code=0,
                    stdout="ok",
                    stderr="",
                    wall_time=1.23,
                    timed_out=False,
                ),
                verify_passed=True,
            ),
            RaceParticipant(
                agent="codex",
                status="stopped",
                total_time=1.50,
                lines_changed=2,
                result=AgentResult(
                    agent="codex",
                    exit_code=130,
                    stdout="",
                    stderr="stopped",
                    wall_time=1.50,
                    timed_out=False,
                ),
                verify_passed=None,
            ),
        ],
        timestamp="2026-03-03T00:00:00+00:00",
    )

    record = _summary_record(summary)
    assert record["winner_agent"] == "claude"
    assert record["winner_time"] == 1.23
    assert len(record["participant_results"]) == 2
    first = record["participant_results"][0]
    assert first["agent"] == "claude"
    assert first["exit_code"] == 0
    assert first["wall_time"] == 1.23


def test_save_race_summary_appends_records(tmp_path: Path) -> None:
    output = tmp_path / ".coderace" / "race-results.json"
    s1 = RaceSummary(
        race_id="r1",
        task_name="t",
        winner_agent="claude",
        winner_time=1.0,
        participants=[],
        timestamp="2026-03-03T00:00:00+00:00",
    )
    s2 = RaceSummary(
        race_id="r2",
        task_name="t",
        winner_agent=None,
        winner_time=None,
        participants=[],
        timestamp="2026-03-03T00:01:00+00:00",
    )
    _save_race_summary(s1, output)
    _save_race_summary(s2, output)
    data = json.loads(output.read_text(encoding="utf-8"))
    assert isinstance(data, list)
    assert [item["race_id"] for item in data] == ["r1", "r2"]


def test_race_cli_no_save_skips_race_results_file(tmp_repo: Path, tmp_path: Path) -> None:
    task_yaml = tmp_path / "race-nosave.yaml"
    task_yaml.write_text(
        f"""name: race-nosave
description: test no-save
repo: {tmp_repo}
test_command: echo ok
agents:
  - claude
""",
        encoding="utf-8",
    )

    summary = RaceSummary(
        race_id="rid",
        task_name="race-nosave",
        winner_agent="claude",
        winner_time=0.5,
        participants=[],
        timestamp="2026-03-03T00:00:00+00:00",
    )

    class FakeLive:
        def __init__(self, *args, **kwargs):
            pass

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc, tb):
            return False

        def update(self, renderable):
            return None

    with patch("coderace.commands.race.has_uncommitted_changes", return_value=False), patch(
        "coderace.commands.race.get_current_ref", return_value="abc12345"
    ), patch("coderace.commands.race.prune_worktrees"), patch(
        "coderace.commands.race.Live", FakeLive
    ), patch("coderace.commands.race.run_race", return_value=summary):
        result = runner.invoke(app, ["race", str(task_yaml), "--no-save"])

    assert result.exit_code == 0
    assert not (tmp_path / ".coderace" / "race-results.json").exists()


def test_race_cli_saves_race_results_file(tmp_repo: Path, tmp_path: Path) -> None:
    task_yaml = tmp_path / "race-save.yaml"
    task_yaml.write_text(
        f"""name: race-save
description: test save
repo: {tmp_repo}
test_command: echo ok
agents:
  - claude
""",
        encoding="utf-8",
    )

    summary = RaceSummary(
        race_id="rid",
        task_name="race-save",
        winner_agent="claude",
        winner_time=0.5,
        participants=[
            RaceParticipant(
                agent="claude",
                status="winner",
                total_time=0.5,
                result=AgentResult(
                    agent="claude",
                    exit_code=0,
                    stdout="",
                    stderr="",
                    wall_time=0.5,
                    timed_out=False,
                ),
            )
        ],
        timestamp="2026-03-03T00:00:00+00:00",
    )

    class FakeLive:
        def __init__(self, *args, **kwargs):
            pass

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc, tb):
            return False

        def update(self, renderable):
            return None

    with patch("coderace.commands.race.has_uncommitted_changes", return_value=False), patch(
        "coderace.commands.race.get_current_ref", return_value="abc12345"
    ), patch("coderace.commands.race.prune_worktrees"), patch(
        "coderace.commands.race.Live", FakeLive
    ), patch("coderace.commands.race.run_race", return_value=summary):
        result = runner.invoke(app, ["race", str(task_yaml)])

    assert result.exit_code == 0
    output = tmp_path / ".coderace" / "race-results.json"
    assert output.exists()
    data = json.loads(output.read_text(encoding="utf-8"))
    assert data[-1]["winner_agent"] == "claude"


def test_resolve_verify_commands_from_verify_command_field(tmp_path: Path) -> None:
    task_yaml = tmp_path / "task.yaml"
    task_yaml.write_text(
        """name: t
description: d
repo: .
test_command: echo ok
verify_command: pytest verify_contract.py -q
agents:
  - claude
""",
        encoding="utf-8",
    )
    task = Task(
        name="t",
        description="d",
        repo=tmp_path,
        test_command="echo ok",
        agents=["claude"],
        verify_command="pytest verify_contract.py -q",
    )
    commands = _resolve_verify_commands(task_yaml, task)
    assert commands == ["pytest verify_contract.py -q"]


def test_resolve_verify_commands_from_verify_string_section(tmp_path: Path) -> None:
    task_yaml = tmp_path / "task.yaml"
    task_yaml.write_text(
        """name: t
description: d
repo: .
test_command: echo ok
verify: pytest verify_a.py -q
agents:
  - claude
""",
        encoding="utf-8",
    )
    task = Task(
        name="t",
        description="d",
        repo=tmp_path,
        test_command="echo ok",
        agents=["claude"],
    )
    commands = _resolve_verify_commands(task_yaml, task)
    assert commands == ["pytest verify_a.py -q"]


def test_resolve_verify_commands_from_verify_list_dedupes(tmp_path: Path) -> None:
    task_yaml = tmp_path / "task.yaml"
    task_yaml.write_text(
        """name: t
description: d
repo: .
test_command: echo ok
verify:
  - pytest verify_a.py -q
  - pytest verify_b.py -q
agents:
  - claude
""",
        encoding="utf-8",
    )
    task = Task(
        name="t",
        description="d",
        repo=tmp_path,
        test_command="echo ok",
        agents=["claude"],
        verify_command="pytest verify_a.py -q",
    )
    commands = _resolve_verify_commands(task_yaml, task)
    assert commands == ["pytest verify_a.py -q", "pytest verify_b.py -q"]


def test_run_race_single_agent_still_works(tmp_repo: Path) -> None:
    task = _make_task(tmp_repo, ["claude"])

    def fake_runner(**kwargs):
        return (
            AgentResult(
                agent="claude",
                exit_code=0,
                stdout="",
                stderr="",
                wall_time=0.01,
                timed_out=False,
            ),
            1,
            {"stopped": False, "total_wall_time": 0.01},
        )

    with patch("coderace.commands.race._invoke_worktree_runner", side_effect=fake_runner):
        summary = run_race(
            task=task,
            base_ref="abc123",
            timeout=60,
            no_cost=True,
            verify_commands=[],
            poll_interval_seconds=0.01,
            refresh_interval_seconds=0.01,
            graceful_shutdown_seconds=0.05,
        )

    assert summary.winner_agent == "claude"
    assert summary.participants[0].status == "winner"


def test_winner_announcement_no_winner_message() -> None:
    summary = RaceSummary(
        race_id="rid",
        task_name="t",
        winner_agent=None,
        winner_time=None,
        participants=[],
        timestamp="2026-03-03T00:00:00+00:00",
    )
    lines = _winner_announcement(summary, verify_commands=[])
    assert lines == ["No winner: all agents failed or timed out."]


def test_winner_announcement_includes_runner_up_delta() -> None:
    summary = RaceSummary(
        race_id="rid",
        task_name="t",
        winner_agent="claude",
        winner_time=1.0,
        participants=[
            RaceParticipant(
                agent="claude",
                status="winner",
                finished_at=1.0,
                total_time=1.0,
                result=AgentResult(
                    agent="claude",
                    exit_code=0,
                    stdout="",
                    stderr="",
                    wall_time=1.0,
                    timed_out=False,
                ),
            ),
            RaceParticipant(
                agent="codex",
                status="stopped",
                finished_at=1.5,
                total_time=1.5,
                result=AgentResult(
                    agent="codex",
                    exit_code=0,
                    stdout="",
                    stderr="",
                    wall_time=1.5,
                    timed_out=False,
                ),
            ),
        ],
        timestamp="2026-03-03T00:00:00+00:00",
    )
    lines = _winner_announcement(summary, verify_commands=["pytest verify.py -q"])
    assert lines[0].startswith("🏆 Winner: claude")
    assert lines[1] == "Runner-up: codex - finished 0:00 later"


def test_build_live_panel_contains_core_labels() -> None:
    panel = _build_live_panel(
        "task-x",
        [
            RaceParticipant(agent="claude", status="coding", started_at=0.0),
            RaceParticipant(agent="codex", status="testing", started_at=0.0),
        ],
    )
    from rich.console import Console

    console = Console(record=True, no_color=True)
    console.print(panel)
    out = console.export_text()
    assert "coderace race - task-x" in out
    assert "Press Ctrl+C to abort" in out
    assert "coding" in out
    assert "testing" in out


def test_race_cli_handles_ctrl_c_and_prunes_worktrees(tmp_repo: Path, tmp_path: Path) -> None:
    task_yaml = tmp_path / "race-interrupt.yaml"
    task_yaml.write_text(
        f"""name: race-interrupt
description: test interrupt
repo: {tmp_repo}
test_command: echo ok
agents:
  - claude
""",
        encoding="utf-8",
    )

    class FakeLive:
        def __init__(self, *args, **kwargs):
            pass

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc, tb):
            return False

        def update(self, renderable):
            return None

    with patch("coderace.commands.race.has_uncommitted_changes", return_value=False), patch(
        "coderace.commands.race.get_current_ref", return_value="abc12345"
    ), patch("coderace.commands.race.Live", FakeLive), patch(
        "coderace.commands.race.run_race", side_effect=KeyboardInterrupt()
    ), patch("coderace.commands.race.prune_worktrees") as prune_mock:
        result = runner.invoke(app, ["race", str(task_yaml), "--no-save"])

    assert result.exit_code == 130
    assert "Race aborted by user." in result.output
    prune_mock.assert_called_once()



"""Codex CLI adapter."""

from __future__ import annotations

from typing import Optional

from coderace.adapters.base import BaseAdapter
from coderace.cost import CostResult, parse_codex_cost

DEFAULT_CODEX_MODEL = "gpt-5.3-codex"


class CodexAdapter(BaseAdapter):
    """Adapter for OpenAI Codex CLI."""

    name = "codex"

    def build_command(self, task_description: str, model: Optional[str] = None) -> list[str]:
        cmd = [
            "codex",
            "exec",
            "--full-auto",
        ]
        effective_model = model or self.model
        if effective_model:
            cmd += ["--model", effective_model]
        cmd.append(task_description)
        return cmd

    def parse_cost(
        self,
        stdout: str,
        stderr: str,
        model_name: str = "",
        custom_pricing: dict[str, tuple[float, float]] | None = None,
    ) -> Optional[CostResult]:
        """Parse cost data from Codex CLI output."""
        effective_model = model_name or self.model or DEFAULT_CODEX_MODEL
        return parse_codex_cost(stdout, stderr, effective_model, custom_pricing)



"""Tests for D4: benchmark and race command model-specific agent handling."""

from __future__ import annotations

from unittest.mock import MagicMock, patch

import pytest

from coderace.adapters import instantiate_adapter, parse_agent_spec


# ---------------------------------------------------------------------------
# Benchmark agent validation: agent:model syntax
# ---------------------------------------------------------------------------


def test_benchmark_agent_validation_with_model_spec() -> None:
    """Benchmark function handles agent:model specs without crashing on unknown agent check."""
    from coderace.adapters import ADAPTERS

    spec = "codex:gpt-5.4"
    agent_base, model = parse_agent_spec(spec)
    assert agent_base in ADAPTERS, f"Base agent {agent_base!r} should be in ADAPTERS"


def test_benchmark_agent_validation_rejects_unknown() -> None:
    from coderace.adapters import ADAPTERS

    spec = "bogus:some-model"
    agent_base, _ = parse_agent_spec(spec)
    assert agent_base not in ADAPTERS


# ---------------------------------------------------------------------------
# TaskAgentResult: agent name is display name (not raw spec)
# ---------------------------------------------------------------------------


def test_instantiate_adapter_name_is_display_name() -> None:
    """The adapter.name set by instantiate_adapter is the display name."""
    adapter = instantiate_adapter("codex:gpt-5.4")
    assert adapter.name == "codex (gpt-5.4)"


def test_instantiate_adapter_no_model_name_is_plain() -> None:
    adapter = instantiate_adapter("claude")
    assert adapter.name == "claude"


# ---------------------------------------------------------------------------
# ELO tracking: model variants are separate entries
# ---------------------------------------------------------------------------


def test_elo_entries_per_model_variant() -> None:
    """Different model variants of the same adapter get separate ELO entries."""
    from coderace.elo import update_pair_ratings

    # Simulate two codex variants — both start at 1500
    rating_winner, rating_loser = 1500.0, 1500.0
    new_winner, new_loser = update_pair_ratings(rating_winner, rating_loser, actual_a=1.0)
    # Winner gains ELO, loser loses
    assert new_winner > 1500.0
    assert new_loser < 1500.0
    # The key insight: since names are distinct strings, store tracks them separately
    ratings = {
        "codex (gpt-5.4)": new_winner,
        "codex (gpt-5.3-codex)": new_loser,
    }
    assert ratings["codex (gpt-5.4)"] != ratings["codex (gpt-5.3-codex)"]


# ---------------------------------------------------------------------------
# Store: model-qualified agent names stored correctly
# ---------------------------------------------------------------------------


def test_store_accepts_model_qualified_agent_names(tmp_path) -> None:
    """Store can save and retrieve results keyed by model-qualified names."""
    import os
    os.environ["CODERACE_DB"] = str(tmp_path / "test.db")
    try:
        from coderace.store import ResultStore
        store = ResultStore()

        # Simulate saving a run with model-qualified agent names
        run_id = store.save_run(
            task_name="fix-bug",
            results=[
                {
                    "agent": "codex (gpt-5.4)",
                    "composite_score": 85.0,
                    "wall_time": 30.0,
                    "lines_changed": 10,
                    "tests_pass": True,
                    "exit_clean": True,
                    "lint_clean": True,
                    "cost_usd": None,
                },
                {
                    "agent": "codex (gpt-5.3-codex)",
                    "composite_score": 72.0,
                    "wall_time": 45.0,
                    "lines_changed": 8,
                    "tests_pass": True,
                    "exit_clean": True,
                    "lint_clean": False,
                    "cost_usd": None,
                },
            ],
        )
        assert run_id is not None

        stats = store.get_agent_stats()
        agent_names = {s.agent for s in stats}
        assert "codex (gpt-5.4)" in agent_names
        assert "codex (gpt-5.3-codex)" in agent_names
    finally:
        del os.environ["CODERACE_DB"]


# ---------------------------------------------------------------------------
# Race command: agent:model syntax validation
# ---------------------------------------------------------------------------


def test_race_command_validates_agent_model_spec() -> None:
    """Race command validation accepts agent:model specs."""
    from coderace.adapters import ADAPTERS

    agents = ["codex:gpt-5.4", "claude:opus-4-6", "bogus:model"]
    valid = [a for a in agents if parse_agent_spec(a)[0] in ADAPTERS]
    assert "codex:gpt-5.4" in valid
    assert "claude:opus-4-6" in valid
    assert "bogus:model" not in valid


# ---------------------------------------------------------------------------
# Leaderboard: model variants appear as separate entries
# ---------------------------------------------------------------------------


def test_leaderboard_shows_model_variants_separately(tmp_path) -> None:
    """Leaderboard lists model variants as distinct entries."""
    import os
    os.environ["CODERACE_DB"] = str(tmp_path / "test.db")
    try:
        from coderace.store import ResultStore
        store = ResultStore()

        store.save_run(
            task_name="task1",
            results=[
                {
                    "agent": "codex (gpt-5.4)",
                    "composite_score": 90.0,
                    "wall_time": 20.0,
                    "lines_changed": 5,
                    "tests_pass": True,
                    "exit_clean": True,
                    "lint_clean": True,
                    "cost_usd": None,
                },
                {
                    "agent": "codex (gpt-5.3-codex)",
                    "composite_score": 75.0,
                    "wall_time": 35.0,
                    "lines_changed": 8,
                    "tests_pass": True,
                    "exit_clean": True,
                    "lint_clean": False,
                    "cost_usd": None,
                },
            ],
        )

        stats = store.get_agent_stats()
        agent_names = [s.agent for s in stats]
        # Both variants should appear
        assert "codex (gpt-5.4)" in agent_names
        assert "codex (gpt-5.3-codex)" in agent_names
        # They should be separate (not merged)
        assert agent_names.count("codex (gpt-5.4)") == 1
        assert agent_names.count("codex (gpt-5.3-codex)") == 1
    finally:
        del os.environ["CODERACE_DB"]



"""Tests for reporter output."""

from __future__ import annotations

import json
from pathlib import Path

from coderace.reporter import load_results_json, print_results, save_results_json
from coderace.types import Score, ScoreBreakdown


def _make_scores() -> list[Score]:
    return [
        Score(
            agent="claude",
            composite=85.0,
            breakdown=ScoreBreakdown(
                tests_pass=True, exit_clean=True, lint_clean=True, wall_time=10.5, lines_changed=42
            ),
        ),
        Score(
            agent="codex",
            composite=70.0,
            breakdown=ScoreBreakdown(
                tests_pass=True,
                exit_clean=True,
                lint_clean=False,
                wall_time=15.2,
                lines_changed=98,
            ),
        ),
    ]


def test_print_results_returns_string() -> None:
    scores = _make_scores()
    output = print_results(scores)
    assert "claude" in output
    assert "codex" in output
    assert "85.0" in output


def test_save_and_load_json(tmp_path: Path) -> None:
    scores = _make_scores()
    json_path = tmp_path / ".coderace" / "results.json"
    save_results_json(scores, json_path)

    assert json_path.exists()
    data = load_results_json(json_path)
    assert len(data) == 2
    assert data[0]["agent"] == "claude"  # Higher score first
    assert data[0]["composite_score"] == 85.0


def test_json_structure(tmp_path: Path) -> None:
    scores = _make_scores()
    json_path = tmp_path / "results.json"
    save_results_json(scores, json_path)

    raw = json.loads(json_path.read_text())
    assert "results" in raw
    entry = raw["results"][0]
    assert "breakdown" in entry
    assert "tests_pass" in entry["breakdown"]
    assert "wall_time" in entry["breakdown"]



"""Tests for statistical aggregation."""

from __future__ import annotations

from coderace.stats import aggregate_runs
from coderace.types import Score, ScoreBreakdown


def _make_score(agent: str, composite: float, wall_time: float = 10.0, lines: int = 50) -> Score:
    return Score(
        agent=agent,
        composite=composite,
        breakdown=ScoreBreakdown(
            tests_pass=composite >= 50,
            exit_clean=True,
            lint_clean=True,
            wall_time=wall_time,
            lines_changed=lines,
        ),
    )


def test_aggregate_single_run() -> None:
    scores = [_make_score("claude", 85.0), _make_score("codex", 70.0)]
    stats = aggregate_runs([scores])
    assert len(stats) == 2
    assert stats[0].agent == "claude"
    assert stats[0].score_mean == 85.0
    assert stats[0].score_stddev == 0.0
    assert stats[0].runs == 1


def test_aggregate_multiple_runs_mean() -> None:
    run1 = [_make_score("claude", 80.0), _make_score("codex", 60.0)]
    run2 = [_make_score("claude", 90.0), _make_score("codex", 70.0)]
    stats = aggregate_runs([run1, run2])
    claude = next(s for s in stats if s.agent == "claude")
    assert claude.score_mean == 85.0
    assert claude.runs == 2


def test_aggregate_multiple_runs_stddev() -> None:
    run1 = [_make_score("claude", 80.0)]
    run2 = [_make_score("claude", 90.0)]
    stats = aggregate_runs([run1, run2])
    claude = stats[0]
    # stddev of [80, 90] = sqrt(((80-85)^2 + (90-85)^2) / 1) = sqrt(50) ~ 7.07
    assert claude.score_stddev > 0


def test_aggregate_sorted_by_mean_desc() -> None:
    run1 = [_make_score("aider", 50.0), _make_score("claude", 90.0), _make_score("codex", 70.0)]
    stats = aggregate_runs([run1])
    assert stats[0].agent == "claude"
    assert stats[1].agent == "codex"
    assert stats[2].agent == "aider"


def test_aggregate_tests_pass_rate() -> None:
    # Run 1: tests pass for claude, Run 2: tests fail (score < 50)
    run1 = [Score("claude", 80.0, ScoreBreakdown(tests_pass=True, exit_clean=True))]
    run2 = [Score("claude", 30.0, ScoreBreakdown(tests_pass=False, exit_clean=True))]
    stats = aggregate_runs([run1, run2])
    assert stats[0].tests_pass_rate == 0.5


def test_aggregate_empty_runs() -> None:
    stats = aggregate_runs([])
    assert stats == []


def test_aggregate_preserves_per_run_scores() -> None:
    run1 = [_make_score("claude", 80.0)]
    run2 = [_make_score("claude", 90.0)]
    stats = aggregate_runs([run1, run2])
    assert sorted(stats[0].per_run_scores) == [80.0, 90.0]



"""CLI command: coderace review."""

from __future__ import annotations

import subprocess
import sys
from pathlib import Path

import typer
from rich.console import Console

from coderace.adapters import ADAPTERS, parse_agent_spec
from coderace.review import DEFAULT_REVIEW_AGENTS, DEFAULT_REVIEW_LANES, run_review
from coderace.review_report import render_review_json, render_review_markdown
from coderace.maintainer_rubric import score_rubric
from coderace.display import MaintainerRubricDisplay

app = typer.Typer(
    help="Run multi-lane parallel agent review on a diff.",
    context_settings={"allow_interspersed_args": True},
)


def _parse_csv_option(value: str) -> list[str]:
    return [item.strip() for item in value.split(",") if item.strip()]


def _validate_agents(agents: list[str]) -> list[str]:
    valid_agents: list[str] = []
    for agent in agents:
        if parse_agent_spec(agent)[0] in ADAPTERS:
            valid_agents.append(agent)
    return valid_agents


def _read_stdin_diff() -> str:
    if sys.stdin.isatty():
        raise ValueError(
            "No diff provided. Use --diff, --commit, --branch, or pipe diff text to stdin."
        )
    return sys.stdin.read()


def _read_diff_source(
    diff_file: Path | None,
    commit: str | None,
    branch: str | None,
    workdir: Path,
) -> str:
    if diff_file is not None:
        if not diff_file.exists():
            raise FileNotFoundError(f"Diff file not found: {diff_file}")
        return diff_file.read_text(encoding="utf-8")
    if commit:
        return _git_diff(["git", "diff", f"{commit}~1", commit], workdir)
    if branch:
        if "..." not in branch:
            raise ValueError("Branch diff must use <base>...<head> syntax.")
        base, head = branch.split("...", 1)
        return _git_diff(["git", "diff", f"{base}...{head}"], workdir)
    return _read_stdin_diff()


def _git_diff(command: list[str], workdir: Path) -> str:
    result = subprocess.run(
        command,
        cwd=workdir,
        capture_output=True,
        text=True,
    )
    if result.returncode != 0:
        raise RuntimeError(result.stderr.strip() or "git diff failed")
    return result.stdout


@app.callback(invoke_without_command=True)
def review_main(
    ctx: typer.Context,
    diff_file: Path | None = typer.Option(
        None,
        "--diff",
        help="Read diff from file (default: stdin)",
    ),
    commit: str | None = typer.Option(
        None,
        "--commit",
        help="Generate diff from commit ref (for example HEAD~1 or abc123)",
    ),
    branch: str | None = typer.Option(
        None,
        "--branch",
        help="Generate diff from branch range (for example main...my-branch)",
    ),
    lanes: str = typer.Option(
        ",".join(DEFAULT_REVIEW_LANES),
        "--lanes",
        help="Comma-separated review lanes",
    ),
    agents: str = typer.Option(
        ",".join(DEFAULT_REVIEW_AGENTS),
        "--agents",
        help="Comma-separated agents",
    ),
    cross_review: bool = typer.Option(
        False,
        "--cross-review",
        help="Run Phase 2 cross-review after Phase 1",
    ),
    output: Path | None = typer.Option(
        None,
        "--output",
        help="Write report to file (default: stdout)",
    ),
    fmt: str = typer.Option(
        "markdown",
        "--format",
        help="Output format: markdown|json",
    ),
    no_color: bool = typer.Option(
        False,
        "--no-color",
        help="Plain output (no rich markup)",
    ),
    maintainer_mode: bool = typer.Option(
        False,
        "--maintainer-mode",
        help="Append maintainer rubric section to review output (static analysis, no LLM)",
    ),
    min_score: int | None = typer.Option(
        None,
        "--min-score",
        help="Minimum maintainer rubric score (0-100). Requires --maintainer-mode. Exit 1 if score < min-score.",
        min=0,
        max=100,
    ),
) -> None:
    """Run multi-lane parallel agent review on a diff."""
    if ctx.invoked_subcommand is not None:
        return

    console = Console(stderr=True, no_color=no_color)
    workdir = Path.cwd()

    try:
        diff_text = _read_diff_source(diff_file, commit, branch, workdir)
    except (FileNotFoundError, RuntimeError, ValueError) as exc:
        console.print(f"[red]{exc}[/red]")
        raise typer.Exit(1)

    if not diff_text.strip():
        console.print("[red]Diff input is empty.[/red]")
        raise typer.Exit(1)

    resolved_lanes = _parse_csv_option(lanes)
    resolved_agents = _parse_csv_option(agents)
    valid_agents = _validate_agents(resolved_agents)
    invalid_agents = [agent for agent in resolved_agents if agent not in valid_agents]
    for agent in invalid_agents:
        console.print(f"[red]Unknown agent: {agent}[/red]")
    if not valid_agents:
        console.print("[red]No valid agents to run.[/red]")
        raise typer.Exit(1)

    try:
        result = run_review(
            diff_text,
            resolved_lanes,
            valid_agents,
            cross_review=cross_review,
            workdir=workdir,
        )
    except ValueError as exc:
        console.print(f"[red]{exc}[/red]")
        raise typer.Exit(1)

    if fmt == "json":
        rendered = render_review_json(result)
    elif fmt == "markdown":
        rendered = render_review_markdown(result)
    else:
        console.print(f"[red]Unknown format: {fmt}. Choose from markdown or json.[/red]")
        raise typer.Exit(1)

    if output is not None:
        output.write_text(rendered, encoding="utf-8")
        console.print(f"[green]Review report written to:[/green] {output}")
        if maintainer_mode:
            exit_code = _append_maintainer_rubric(diff_text, console, no_color, min_score=min_score)
            raise typer.Exit(exit_code)
        return

    typer.echo(rendered, nl=False)

    if maintainer_mode:
        exit_code = _append_maintainer_rubric(diff_text, console, no_color, min_score=min_score)
        raise typer.Exit(exit_code)


def _append_maintainer_rubric(
    diff_text: str,
    console: Console,
    no_color: bool,
    min_score: int | None = None,
) -> int:
    """Score the diff with the maintainer rubric and print the result.

    Returns exit code: 0 (pass) or 1 (fail, only when min_score is set).
    """
    rubric = score_rubric(diff_text)
    display_console = Console(no_color=no_color)
    display = MaintainerRubricDisplay()
    display_console.print()
    display.print(rubric, console=display_console)

    if min_score is not None:
        score_int = round(rubric.composite)
        if rubric.composite >= min_score:
            display_console.print(
                f"✅ Maintainer score {score_int} ≥ {min_score} (gate: PASS)"
            )
            return 0
        else:
            display_console.print(
                f"❌ Maintainer score {score_int} < {min_score} (gate: FAIL)"
            )
            return 1

    return 0



"""Tests for the GitHub Action review mode infrastructure.

Covers:
  - scripts/format-review-comment.py
  - action.yml structure (new inputs/outputs)
  - Example workflow YAML validity
  - Integration smoke test: empty diff case
"""

from __future__ import annotations

import importlib.util
import json
from pathlib import Path

import pytest

# ---------------------------------------------------------------------------
# Load format-review-comment.py via importlib (hyphenated filename)
# ---------------------------------------------------------------------------

_SCRIPTS_DIR = Path(__file__).parent.parent / "scripts"
_REPO_ROOT = Path(__file__).parent.parent


def _load_format_review_comment():
    spec = importlib.util.spec_from_file_location(
        "format_review_comment",
        _SCRIPTS_DIR / "format-review-comment.py",
    )
    assert spec is not None and spec.loader is not None
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)  # type: ignore[union-attr]
    return module


frc = _load_format_review_comment()


# ---------------------------------------------------------------------------
# Sample data matching ReviewResult schema (coderace/types.py)
# ---------------------------------------------------------------------------

SAMPLE_REVIEW_DATA = {
    "diff_summary": {"files": ["api/handlers.py", "api/utils.py"], "added": 25, "removed": 10},
    "lanes": ["security", "logic"],
    "phase1_findings": [
        {
            "lane": "security",
            "agent": "claude",
            "severity": "critical",
            "finding": "User input used directly in SQL query — SQL injection risk.",
            "location": "api/handlers.py:42",
        },
        {
            "lane": "security",
            "agent": "claude",
            "severity": "warning",
            "finding": "Missing authentication check before sensitive operation.",
            "location": "api/handlers.py:67",
        },
        {
            "lane": "logic",
            "agent": "codex",
            "severity": "error",
            "finding": "Off-by-one error in pagination logic.",
            "location": "api/utils.py:13",
        },
    ],
    "phase2_findings": [
        {
            "lane": "cross-review",
            "agent": "claude",
            "severity": "critical",
            "finding": "SQL injection confirmed — no parameterized queries used.",
            "location": "api/handlers.py:42",
        },
    ],
    "agents_used": ["claude", "codex"],
    "elapsed_seconds": 12.4,
    "timestamp": "2026-03-10T15:00:00",
}

EMPTY_REVIEW_DATA = {
    "diff_summary": {"files": [], "added": 0, "removed": 0},
    "lanes": ["security", "logic"],
    "phase1_findings": [],
    "phase2_findings": [],
    "agents_used": ["claude"],
    "elapsed_seconds": 5.1,
    "timestamp": "2026-03-10T15:00:00",
}


# ---------------------------------------------------------------------------
# format_header
# ---------------------------------------------------------------------------


def test_format_header_counts() -> None:
    header = frc.format_header(SAMPLE_REVIEW_DATA)
    assert "2 agent(s)" in header
    assert "2 lane(s)" in header
    assert "12.4s" in header


def test_format_header_minimal() -> None:
    data = {"agents_used": [], "lanes": [], "elapsed_seconds": 0.0}
    header = frc.format_header(data)
    assert "## coderace review" in header


# ---------------------------------------------------------------------------
# format_summary
# ---------------------------------------------------------------------------


def test_format_summary_with_findings() -> None:
    summary = frc.format_summary(SAMPLE_REVIEW_DATA)
    # Total = 3 phase1 + 1 phase2 = 4
    assert "4 finding(s)" in summary


def test_format_summary_no_findings() -> None:
    summary = frc.format_summary(EMPTY_REVIEW_DATA)
    assert "no issues" in summary.lower()
    assert "✅" in summary


def test_format_summary_top_critical() -> None:
    summary = frc.format_summary(SAMPLE_REVIEW_DATA)
    assert "SQL injection" in summary or "api/handlers.py:42" in summary


# ---------------------------------------------------------------------------
# format_lane_section
# ---------------------------------------------------------------------------


def test_format_lane_section_contains_lanes() -> None:
    section = frc.format_lane_section(SAMPLE_REVIEW_DATA)
    assert "Security" in section
    assert "Logic" in section


def test_format_lane_section_agent_attribution() -> None:
    section = frc.format_lane_section(SAMPLE_REVIEW_DATA)
    assert "claude" in section
    assert "codex" in section


def test_format_lane_section_findings() -> None:
    section = frc.format_lane_section(SAMPLE_REVIEW_DATA)
    assert "SQL injection" in section
    assert "api/handlers.py:42" in section


def test_format_lane_section_empty_lanes() -> None:
    section = frc.format_lane_section(EMPTY_REVIEW_DATA)
    assert "No findings" in section or "no findings" in section.lower()


# ---------------------------------------------------------------------------
# format_cross_review_section
# ---------------------------------------------------------------------------


def test_format_cross_review_section_present() -> None:
    section = frc.format_cross_review_section(SAMPLE_REVIEW_DATA)
    assert "Phase 2" in section
    assert "claude" in section


def test_format_cross_review_section_empty() -> None:
    section = frc.format_cross_review_section(EMPTY_REVIEW_DATA)
    assert section == ""


# ---------------------------------------------------------------------------
# format_review_comment (full output)
# ---------------------------------------------------------------------------


def test_format_review_comment_marker() -> None:
    comment = frc.format_review_comment(SAMPLE_REVIEW_DATA)
    assert "<!-- coderace-review -->" in comment


def test_format_review_comment_includes_header() -> None:
    comment = frc.format_review_comment(SAMPLE_REVIEW_DATA)
    assert "## coderace review" in comment


def test_format_review_comment_includes_summary() -> None:
    comment = frc.format_review_comment(SAMPLE_REVIEW_DATA)
    assert "finding(s)" in comment


def test_format_review_comment_includes_json_collapsible() -> None:
    raw = json.dumps(SAMPLE_REVIEW_DATA)
    comment = frc.format_review_comment(SAMPLE_REVIEW_DATA, json_raw=raw)
    assert "<details>" in comment
    assert "```json" in comment


def test_format_review_comment_no_issues() -> None:
    comment = frc.format_review_comment(EMPTY_REVIEW_DATA)
    assert "<!-- coderace-review -->" in comment
    assert "no issues" in comment.lower()


def test_format_review_comment_footer() -> None:
    comment = frc.format_review_comment(SAMPLE_REVIEW_DATA)
    assert "coderace" in comment.lower()


# ---------------------------------------------------------------------------
# format_empty_comment
# ---------------------------------------------------------------------------


def test_format_empty_comment_has_marker() -> None:
    comment = frc.format_empty_comment()
    assert "<!-- coderace-review -->" in comment


def test_format_empty_comment_custom_reason() -> None:
    comment = frc.format_empty_comment("JSON file not found.")
    assert "JSON file not found." in comment


# ---------------------------------------------------------------------------
# main() CLI entry point
# ---------------------------------------------------------------------------


def test_main_missing_json(tmp_path: Path, capsys: pytest.CaptureFixture) -> None:
    missing = tmp_path / "nonexistent-review.json"
    ret = frc.main(["--json-file", str(missing)])
    assert ret == 0
    captured = capsys.readouterr()
    assert "<!-- coderace-review -->" in captured.out


def test_main_writes_to_file(tmp_path: Path) -> None:
    json_file = tmp_path / "review.json"
    json_file.write_text(json.dumps(SAMPLE_REVIEW_DATA))
    out_file = tmp_path / "comment.md"

    ret = frc.main(["--json-file", str(json_file), "--output", str(out_file)])
    assert ret == 0
    assert out_file.exists()
    content = out_file.read_text()
    assert "<!-- coderace-review -->" in content
    assert "## coderace review" in content


def test_main_invalid_json(tmp_path: Path, capsys: pytest.CaptureFixture) -> None:
    bad = tmp_path / "bad-review.json"
    bad.write_text("NOT JSON AT ALL")
    ret = frc.main(["--json-file", str(bad)])
    assert ret == 0
    captured = capsys.readouterr()
    assert "<!-- coderace-review -->" in captured.out


def test_main_empty_json(tmp_path: Path, capsys: pytest.CaptureFixture) -> None:
    empty = tmp_path / "empty.json"
    empty.write_text("")
    ret = frc.main(["--json-file", str(empty)])
    assert ret == 0
    captured = capsys.readouterr()
    assert "<!-- coderace-review -->" in captured.out


def test_main_stdout_default(tmp_path: Path, capsys: pytest.CaptureFixture) -> None:
    json_file = tmp_path / "review.json"
    json_file.write_text(json.dumps(EMPTY_REVIEW_DATA))
    ret = frc.main(["--json-file", str(json_file)])
    assert ret == 0
    captured = capsys.readouterr()
    assert "<!-- coderace-review -->" in captured.out


# ---------------------------------------------------------------------------
# action.yml structure checks
# ---------------------------------------------------------------------------


def test_action_yml_has_mode_input() -> None:
    """action.yml must have the new mode input."""
    action_yml = (_REPO_ROOT / "action.yml").read_text()
    assert "mode:" in action_yml
    assert "review" in action_yml


def test_action_yml_has_diff_source_input() -> None:
    action_yml = (_REPO_ROOT / "action.yml").read_text()
    assert "diff-source:" in action_yml


def test_action_yml_has_review_outputs() -> None:
    action_yml = (_REPO_ROOT / "action.yml").read_text()
    assert "review-json" in action_yml
    assert "review-md" in action_yml


def test_action_yml_backward_compatible() -> None:
    """mode: run path still references ci-run.sh."""
    action_yml = (_REPO_ROOT / "action.yml").read_text()
    assert "ci-run.sh" in action_yml


def test_action_yml_review_path() -> None:
    """mode: review path references ci-review.sh."""
    action_yml = (_REPO_ROOT / "action.yml").read_text()
    assert "ci-review.sh" in action_yml


# ---------------------------------------------------------------------------
# ci-review.sh existence + executable
# ---------------------------------------------------------------------------


def test_ci_review_sh_exists() -> None:
    script = _REPO_ROOT / "scripts" / "ci-review.sh"
    assert script.exists(), "scripts/ci-review.sh must exist"


def test_ci_review_sh_is_executable() -> None:
    import stat

    script = _REPO_ROOT / "scripts" / "ci-review.sh"
    mode = script.stat().st_mode
    assert mode & stat.S_IXUSR, "scripts/ci-review.sh must be executable"


# ---------------------------------------------------------------------------
# Example workflow YAML validity
# ---------------------------------------------------------------------------


def test_example_workflow_is_valid_yaml() -> None:
    """Example coderace-pr-review.yml must be valid YAML."""
    example = _REPO_ROOT / ".github" / "workflows" / "examples" / "coderace-pr-review.yml"
    assert example.exists(), "Example workflow file must exist"
    try:
        import yaml  # type: ignore[import]
    except ImportError:
        pytest.skip("pyyaml not installed")
    data = yaml.safe_load(example.read_text())
    assert data is not None
    assert "on" in data or True  # YAML key "on" may parse as True
    assert "jobs" in data


def test_example_workflow_uses_mode_review() -> None:
    example = _REPO_ROOT / ".github" / "workflows" / "examples" / "coderace-pr-review.yml"
    content = example.read_text()
    assert "mode: review" in content


def test_example_workflow_fetch_depth_0() -> None:
    """Must use fetch-depth: 0 for git diff to work."""
    example = _REPO_ROOT / ".github" / "workflows" / "examples" / "coderace-pr-review.yml"
    content = example.read_text()
    assert "fetch-depth: 0" in content


# ---------------------------------------------------------------------------
# Integration smoke test: empty diff path via coderace review
# ---------------------------------------------------------------------------


def test_review_command_exists() -> None:
    """coderace review subcommand must be importable and registered."""
    from coderace.cli import app  # noqa: F401
    from coderace.commands.review import app as review_app  # noqa: F401

    assert review_app is not None


def test_review_render_empty_result() -> None:
    """render_review_markdown and render_review_json should not crash on minimal input."""
    from coderace.review_report import render_review_json, render_review_markdown
    from coderace.types import ReviewResult

    result = ReviewResult(
        diff_summary={"files": [], "added": 0, "removed": 0},
        lanes=[],
        phase1_findings=[],
        phase2_findings=[],
        agents_used=[],
        elapsed_seconds=0.0,
        timestamp="2026-03-10T15:00:00",
    )
    md = render_review_markdown(result)
    assert isinstance(md, str)
    js = render_review_json(result)
    assert isinstance(js, str)
    data = json.loads(js)
    assert "lanes" in data



"""Benchmark report rendering: terminal, markdown, and HTML."""

from __future__ import annotations

import html as _html_mod
from datetime import datetime, timezone
from typing import Optional

from rich.console import Console
from rich.table import Table

from coderace.benchmark import BenchmarkResult
from coderace.benchmark_stats import BenchmarkStats
from coderace.statistics import compute_aggregate_stats, compute_trial_stats


def _has_verification(result: BenchmarkResult) -> bool:
    return any(r.verify_applicable for r in result.results)


def _task_verify_percentage(
    result: BenchmarkResult,
    task_name: str,
    lookup: dict[tuple[str, str], object],
) -> str:
    applicable = 0
    passed = 0
    for agent in result.agents:
        entry = lookup.get((task_name, agent))
        if not entry or not entry.verify_applicable:
            continue
        applicable += 1
        if entry.verify_passed:
            passed += 1
    if applicable == 0:
        return "-"
    pct = int(round(passed / applicable * 100))
    return f"{pct}%"


def _truncate_output(output: str, max_lines: int = 20) -> str:
    if not output:
        return ""
    lines = output.splitlines()
    if len(lines) <= max_lines:
        return "\n".join(lines)
    hidden = len(lines) - max_lines
    return "\n".join(lines[:max_lines] + [f"... (+{hidden} more lines)"])


def _verify_details_rows(
    result: BenchmarkResult,
    lookup: dict[tuple[str, str], object],
) -> list[tuple[str, str, str, str]]:
    rows: list[tuple[str, str, str, str]] = []
    for task_name in result.tasks:
        for agent in result.agents:
            entry = lookup.get((task_name, agent))
            if not entry or not entry.verify_applicable:
                continue
            verify_pct = f"{int(round(entry.verify_score))}%"
            rows.append(
                (
                    task_name,
                    agent,
                    verify_pct,
                    _truncate_output(entry.verify_output),
                )
            )
    return rows


def _is_multi_trial(result: BenchmarkResult) -> bool:
    if result.trials > 1:
        return True
    seen: set[tuple[str, str]] = set()
    for row in result.results:
        key = (row.task_name, row.agent)
        if key in seen:
            return True
        seen.add(key)
    return False


def _task_reliability(result: BenchmarkResult, task_name: str) -> float:
    task_rows = [row for row in result.results if row.task_name == task_name]
    if not task_rows:
        return 0.0
    ok = sum(1 for row in task_rows if not row.timed_out and not row.error)
    return ok / len(task_rows)


def _render_elo_terminal(console: Console, elo_ratings: dict[str, float]) -> None:
    if not elo_ratings:
        return
    table = Table(title="ELO Ratings", show_lines=True, expand=False)
    table.add_column("Rank", justify="right")
    table.add_column("Agent", style="cyan")
    table.add_column("Rating", justify="right")
    for idx, (agent, rating) in enumerate(
        sorted(elo_ratings.items(), key=lambda item: item[1], reverse=True),
        start=1,
    ):
        table.add_row(str(idx), agent, f"{rating:.1f}")
    console.print()
    console.print(table)


def _render_elo_markdown(elo_ratings: dict[str, float]) -> str:
    if not elo_ratings:
        return ""
    lines = ["## ELO Ratings", "", "| Rank | Agent | Rating |", "|------|-------|--------|"]
    for idx, (agent, rating) in enumerate(
        sorted(elo_ratings.items(), key=lambda item: item[1], reverse=True),
        start=1,
    ):
        lines.append(f"| {idx} | {agent} | {rating:.1f} |")
    lines.append("")
    return "\n".join(lines)


def _render_elo_html(elo_ratings: dict[str, float]) -> str:
    if not elo_ratings:
        return ""
    rows = []
    for idx, (agent, rating) in enumerate(
        sorted(elo_ratings.items(), key=lambda item: item[1], reverse=True),
        start=1,
    ):
        rows.append(
            "<tr>"
            f"<td>{idx}</td>"
            f"<td>{_html_mod.escape(agent)}</td>"
            f"<td>{rating:.1f}</td>"
            "</tr>"
        )
    rows_html = "\n".join(rows)
    return f"""
<h2>ELO Ratings</h2>
<table>
<thead><tr><th>Rank</th><th>Agent</th><th>Rating</th></tr></thead>
<tbody>
{rows_html}
</tbody>
</table>
"""


def _render_benchmark_terminal_trials(
    result: BenchmarkResult,
    console: Console,
    elo_ratings: dict[str, float] | None = None,
) -> None:
    trial_stats = compute_trial_stats(result)
    aggregate_stats = compute_aggregate_stats(result, trial_stats=trial_stats)
    trial_lookup = {(row.task_name, row.agent): row for row in trial_stats}
    aggregate_lookup = {row.agent: row for row in aggregate_stats}

    table = Table(title="coderace benchmark (statistical)", show_lines=True, expand=False)
    table.add_column("Task", style="bold")
    for agent in result.agents:
        table.add_column(agent, justify="center")
    table.add_column("CI (95%)", justify="center")
    table.add_column("Consistency", justify="center")
    table.add_column("Reliability", justify="center")

    for task_name in result.tasks:
        row_values = [task_name]
        task_stat_rows = []
        for agent in result.agents:
            stat = trial_lookup.get((task_name, agent))
            if stat is None:
                row_values.append("-")
                continue
            task_stat_rows.append(stat)
            row_values.append(
                f"[green]{stat.mean_score:.1f}+/-{stat.stddev_score:.1f}[/green] "
                f"[dim]({stat.mean_wall_time:.0f}s)[/dim]"
            )
        if task_stat_rows:
            best = max(task_stat_rows, key=lambda r: r.mean_score)
            ci_text = f"{best.ci_95[0]:.1f}..{best.ci_95[1]:.1f}"
            consistency = sum(r.consistency_score for r in task_stat_rows) / len(task_stat_rows)
        else:
            ci_text = "-"
            consistency = 0.0
        reliability = _task_reliability(result, task_name)
        row_values.extend(
            [
                ci_text,
                f"{int(round(consistency * 100))}%",
                f"{int(round(reliability * 100))}%",
            ]
        )
        table.add_row(*row_values)

    mean_row = ["[bold]Mean Score[/bold]"]
    for agent in result.agents:
        agg = aggregate_lookup.get(agent)
        mean_row.append(f"[bold]{agg.mean_score:.1f}[/bold]" if agg else "-")
    mean_row.extend(["-", "-", "-"])
    table.add_row(*mean_row)

    ci_row = ["[bold]CI (95%)[/bold]"]
    for agent in result.agents:
        agg = aggregate_lookup.get(agent)
        if agg:
            ci_row.append(f"{agg.score_ci_95[0]:.1f}..{agg.score_ci_95[1]:.1f}")
        else:
            ci_row.append("-")
    ci_row.extend(["-", "-", "-"])
    table.add_row(*ci_row)

    win_row = ["[bold]Win Rate[/bold]"]
    for agent in result.agents:
        agg = aggregate_lookup.get(agent)
        win_row.append(f"{int(round((agg.win_rate if agg else 0.0) * 100))}%")
    win_row.extend(["-", "-", "-"])
    table.add_row(*win_row)

    rel_row = ["[bold]Reliability[/bold]"]
    for agent in result.agents:
        agg = aggregate_lookup.get(agent)
        rel_row.append(f"{int(round((agg.reliability if agg else 0.0) * 100))}%")
    rel_row.extend(["-", "-", "-"])
    table.add_row(*rel_row)

    console.print(table)

    if aggregate_stats:
        winner = aggregate_stats[0]
        console.print(
            f"\n[bold green]Winner: {winner.agent}[/bold green] "
            f"(mean score: {winner.mean_score:.1f})"
        )

    _render_elo_terminal(console, elo_ratings or {})


def _render_benchmark_markdown_trials(
    result: BenchmarkResult,
    elo_ratings: dict[str, float] | None = None,
) -> str:
    trial_stats = compute_trial_stats(result)
    aggregate_stats = compute_aggregate_stats(result, trial_stats=trial_stats)
    trial_lookup = {(row.task_name, row.agent): row for row in trial_stats}
    aggregate_lookup = {row.agent: row for row in aggregate_stats}

    lines: list[str] = []
    lines.append("# coderace Benchmark Results\n")
    lines.append(f"Generated: {datetime.now(timezone.utc).strftime('%Y-%m-%d %H:%M UTC')}\n")
    lines.append(f"Benchmark ID: `{result.benchmark_id}`\n")
    lines.append("")

    headers = ["Task", *result.agents, "CI (95%)", "Consistency", "Reliability"]
    separator = "|" + "|".join(["------"] * len(headers)) + "|"
    lines.append("| " + " | ".join(headers) + " |")
    lines.append(separator)

    for task_name in result.tasks:
        cells = [task_name]
        task_rows = []
        for agent in result.agents:
            stat = trial_lookup.get((task_name, agent))
            if stat is None:
                cells.append("-")
                continue
            task_rows.append(stat)
            cells.append(f"{stat.mean_score:.1f}+/-{stat.stddev_score:.1f} ({stat.mean_wall_time:.0f}s)")
        if task_rows:
            best = max(task_rows, key=lambda r: r.mean_score)
            ci_text = f"{best.ci_95[0]:.1f}..{best.ci_95[1]:.1f}"
            consistency = sum(r.consistency_score for r in task_rows) / len(task_rows)
        else:
            ci_text = "-"
            consistency = 0.0
        reliability = _task_reliability(result, task_name)
        cells.append(ci_text)
        cells.append(f"{int(round(consistency * 100))}%")
        cells.append(f"{int(round(reliability * 100))}%")
        lines.append("| " + " | ".join(cells) + " |")

    lines.append(separator)

    mean_cells = ["**Mean Score**"]
    for agent in result.agents:
        agg = aggregate_lookup.get(agent)
        mean_cells.append(f"**{agg.mean_score:.1f}**" if agg else "-")
    mean_cells.extend(["-", "-", "-"])
    lines.append("| " + " | ".join(mean_cells) + " |")

    ci_cells = ["**CI (95%)**"]
    for agent in result.agents:
        agg = aggregate_lookup.get(agent)
        ci_cells.append(f"{agg.score_ci_95[0]:.1f}..{agg.score_ci_95[1]:.1f}" if agg else "-")
    ci_cells.extend(["-", "-", "-"])
    lines.append("| " + " | ".join(ci_cells) + " |")

    win_cells = ["**Win Rate**"]
    for agent in result.agents:
        agg = aggregate_lookup.get(agent)
        win_cells.append(f"{int(round((agg.win_rate if agg else 0.0) * 100))}%")
    win_cells.extend(["-", "-", "-"])
    lines.append("| " + " | ".join(win_cells) + " |")

    rel_cells = ["**Reliability**"]
    for agent in result.agents:
        agg = aggregate_lookup.get(agent)
        rel_cells.append(f"{int(round((agg.reliability if agg else 0.0) * 100))}%")
    rel_cells.extend(["-", "-", "-"])
    lines.append("| " + " | ".join(rel_cells) + " |")

    lines.append("")
    elo_section = _render_elo_markdown(elo_ratings or {})
    if elo_section:
        lines.append(elo_section)
    return "\n".join(lines)


def _render_benchmark_html_trials(
    result: BenchmarkResult,
    elo_ratings: dict[str, float] | None = None,
) -> str:
    trial_stats = compute_trial_stats(result)
    aggregate_stats = compute_aggregate_stats(result, trial_stats=trial_stats)
    trial_lookup = {(row.task_name, row.agent): row for row in trial_stats}
    aggregate_lookup = {row.agent: row for row in aggregate_stats}
    now = datetime.now(timezone.utc).strftime("%Y-%m-%d %H:%M UTC")

    def esc(s: str) -> str:
        return _html_mod.escape(str(s))

    task_rows = ""
    for task_name in result.tasks:
        task_rows += f"<tr><td class='task'>{esc(task_name)}</td>"
        task_stats = []
        for agent in result.agents:
            stat = trial_lookup.get((task_name, agent))
            if stat is None:
                task_rows += "<td>-</td>"
            else:
                task_stats.append(stat)
                task_rows += (
                    f"<td class='pass'>{stat.mean_score:.1f}+/-{stat.stddev_score:.1f}"
                    f"<br><small>({stat.mean_wall_time:.0f}s)</small></td>"
                )
        if task_stats:
            best = max(task_stats, key=lambda row: row.mean_score)
            ci_text = f"{best.ci_95[0]:.1f}..{best.ci_95[1]:.1f}"
            consistency = sum(row.consistency_score for row in task_stats) / len(task_stats)
        else:
            ci_text = "-"
            consistency = 0.0
        reliability = _task_reliability(result, task_name)
        task_rows += f"<td>{esc(ci_text)}</td>"
        task_rows += f"<td>{int(round(consistency * 100))}%</td>"
        task_rows += f"<td>{int(round(reliability * 100))}%</td>"
        task_rows += "</tr>\n"

    mean_cells = ""
    ci_cells = ""
    win_cells = ""
    rel_cells = ""
    for agent in result.agents:
        agg = aggregate_lookup.get(agent)
        if agg:
            mean_cells += f"<td><strong>{agg.mean_score:.1f}</strong></td>"
            ci_cells += f"<td>{agg.score_ci_95[0]:.1f}..{agg.score_ci_95[1]:.1f}</td>"
            win_cells += f"<td>{int(round(agg.win_rate * 100))}%</td>"
            rel_cells += f"<td>{int(round(agg.reliability * 100))}%</td>"
        else:
            mean_cells += "<td>-</td>"
            ci_cells += "<td>-</td>"
            win_cells += "<td>-</td>"
            rel_cells += "<td>-</td>"
    winner = aggregate_stats[0].agent if aggregate_stats else "-"
    agent_headers = "".join(f"<th>{esc(agent)}</th>" for agent in result.agents)
    elo_section = _render_elo_html(elo_ratings or {})

    return f"""<!DOCTYPE html>
<html lang="en">
<head>
<meta charset="UTF-8">
<meta name="viewport" content="width=device-width, initial-scale=1.0">
<title>coderace Benchmark — {esc(result.benchmark_id)}</title>
<style>
  body {{ font-family: system-ui, sans-serif; background: #0d1117; color: #c9d1d9; margin: 0; padding: 2rem; }}
  h1, h2 {{ color: #58a6ff; }}
  .meta {{ color: #8b949e; font-size: 0.9rem; margin-bottom: 1.5rem; }}
  .winner {{ background: #1f6feb33; border: 1px solid #1f6feb; border-radius: 6px; padding: 0.75rem 1rem; margin-bottom: 1.5rem; }}
  table {{ border-collapse: collapse; width: 100%; margin-bottom: 2rem; }}
  th {{ background: #161b22; color: #58a6ff; padding: 0.6rem 1rem; border: 1px solid #30363d; }}
  td {{ padding: 0.5rem 1rem; border: 1px solid #30363d; text-align: center; }}
  td.task {{ text-align: left; font-weight: 500; color: #e6edf3; }}
  td.pass {{ color: #3fb950; }}
  tr.summary td {{ background: #161b22; font-weight: bold; color: #e6edf3; }}
  small {{ color: #8b949e; }}
</style>
</head>
<body>
<h1>coderace Benchmark (Statistical)</h1>
<div class="meta">
  ID: {esc(result.benchmark_id)} &nbsp;|&nbsp; Generated: {now}
</div>
<div class="winner">
  <strong>Winner:</strong> {esc(winner)}
</div>
<h2>Results</h2>
<table>
<thead><tr><th>Task</th>{agent_headers}<th>CI (95%)</th><th>Consistency</th><th>Reliability</th></tr></thead>
<tbody>
{task_rows}
<tr class="summary"><td>Mean Score</td>{mean_cells}<td>-</td><td>-</td><td>-</td></tr>
<tr class="summary"><td>CI (95%)</td>{ci_cells}<td>-</td><td>-</td><td>-</td></tr>
<tr class="summary"><td>Win Rate</td>{win_cells}<td>-</td><td>-</td><td>-</td></tr>
<tr class="summary"><td>Reliability</td>{rel_cells}<td>-</td><td>-</td><td>-</td></tr>
</tbody>
</table>
{elo_section}
</body>
</html>
"""


def render_benchmark_terminal(
    result: BenchmarkResult,
    stats: BenchmarkStats,
    console: Optional[Console] = None,
    elo_ratings: dict[str, float] | None = None,
) -> None:
    """Render a Rich terminal table for the benchmark results."""
    console = console or Console()
    if _is_multi_trial(result):
        _render_benchmark_terminal_trials(result, console, elo_ratings=elo_ratings)
        return

    agents = result.agents
    tasks = result.tasks
    has_verify = _has_verification(result)

    # Lookup: (task, agent) -> TaskAgentResult
    lookup = {(r.task_name, r.agent): r for r in result.results}

    # Build table: tasks as rows, agents as columns
    table = Table(title="coderace benchmark", show_lines=True, expand=False)
    table.add_column("Task", style="bold")
    for agent in agents:
        table.add_column(agent, justify="center")
    if has_verify:
        table.add_column("Verify", justify="center")

    for task_name in tasks:
        row = [task_name]
        for agent in agents:
            r = lookup.get((task_name, agent))
            if r is None:
                row.append("-")
            elif r.error:
                row.append("[red]ERR[/red]")
            elif r.timed_out:
                row.append(f"[yellow]TIMEOUT[/yellow]")
            else:
                score_str = f"{r.score:.1f}"
                time_str = f"({r.wall_time:.0f}s)"
                row.append(f"[green]{score_str}[/green] [dim]{time_str}[/dim]")
        if has_verify:
            row.append(_task_verify_percentage(result, task_name, lookup))
        table.add_row(*row)

    # Summary rows
    agent_stat_map = {s.agent: s for s in stats.agent_stats}

    # Total score row
    total_row = ["[bold]TOTAL[/bold]"]
    for agent in agents:
        s = agent_stat_map.get(agent)
        total_row.append(f"[bold]{s.total_score:.1f}[/bold]" if s else "-")
    if has_verify:
        total_row.append("-")
    table.add_row(*total_row)

    # Win rate row
    win_row = ["[bold]Win Rate[/bold]"]
    for agent in agents:
        s = agent_stat_map.get(agent)
        if s and s.task_count > 0:
            pct = int(s.win_count / s.task_count * 100)
            win_row.append(f"{pct}%")
        else:
            win_row.append("-")
    if has_verify:
        win_row.append("-")
    table.add_row(*win_row)

    # Avg time row
    time_row = ["[bold]Avg Time[/bold]"]
    for agent in agents:
        s = agent_stat_map.get(agent)
        time_row.append(f"{s.avg_time:.1f}s" if s else "-")
    if has_verify:
        time_row.append("-")
    table.add_row(*time_row)

    # Total cost row
    cost_row = ["[bold]Total Cost[/bold]"]
    for agent in agents:
        s = agent_stat_map.get(agent)
        if s and s.total_cost is not None:
            cost_row.append(f"${s.total_cost:.4f}")
        else:
            cost_row.append("-")
    if has_verify:
        cost_row.append("-")
    table.add_row(*cost_row)

    console.print(table)

    # Winner callout
    if stats.agent_stats:
        winner = stats.agent_stats[0]
        console.print(f"\n[bold green]Winner: {winner.agent}[/bold green] "
                      f"(total score: {winner.total_score:.1f}, "
                      f"{winner.win_count}/{winner.task_count} task wins)")

    if has_verify:
        details = _verify_details_rows(result, lookup)
        if details:
            verify_table = Table(title="verification details", show_lines=True, expand=False)
            verify_table.add_column("Task", style="bold")
            verify_table.add_column("Agent", style="cyan")
            verify_table.add_column("Verify", justify="center")
            verify_table.add_column("Output")
            for task_name, agent, verify_pct, output in details:
                verify_table.add_row(task_name, agent, verify_pct, output or "-")
            console.print()
            console.print(verify_table)

    _render_elo_terminal(console, elo_ratings or {})


def render_benchmark_markdown(
    result: BenchmarkResult,
    stats: BenchmarkStats,
    elo_ratings: dict[str, float] | None = None,
) -> str:
    """Render a GitHub-flavored markdown table for the benchmark results."""
    if _is_multi_trial(result):
        return _render_benchmark_markdown_trials(result, elo_ratings=elo_ratings)
    agents = result.agents
    tasks = result.tasks
    lookup = {(r.task_name, r.agent): r for r in result.results}
    agent_stat_map = {s.agent: s for s in stats.agent_stats}
    has_verify = _has_verification(result)

    lines: list[str] = []
    lines.append("# coderace Benchmark Results\n")
    lines.append(f"Generated: {datetime.now(timezone.utc).strftime('%Y-%m-%d %H:%M UTC')}\n")
    lines.append(f"Benchmark ID: `{result.benchmark_id}`\n")

    # Header row
    header_cells = ["Task", *agents]
    if has_verify:
        header_cells.append("Verify")
    header = "| " + " | ".join(header_cells) + " |"
    separator = "|" + "|".join(["------"] * len(header_cells)) + "|"
    lines.append(header)
    lines.append(separator)

    for task_name in tasks:
        cells = [task_name]
        for agent in agents:
            r = lookup.get((task_name, agent))
            if r is None:
                cells.append("-")
            elif r.error:
                cells.append("ERR")
            elif r.timed_out:
                cells.append("TIMEOUT")
            else:
                cells.append(f"{r.score:.1f} ({r.wall_time:.0f}s)")
        if has_verify:
            cells.append(_task_verify_percentage(result, task_name, lookup))
        lines.append("| " + " | ".join(cells) + " |")

    # Summary separator
    lines.append(separator)

    # Total
    total_cells = ["**TOTAL**"]
    for agent in agents:
        s = agent_stat_map.get(agent)
        total_cells.append(f"**{s.total_score:.1f}**" if s else "-")
    if has_verify:
        total_cells.append("-")
    lines.append("| " + " | ".join(total_cells) + " |")

    # Win rate
    win_cells = ["**Win Rate**"]
    for agent in agents:
        s = agent_stat_map.get(agent)
        if s and s.task_count > 0:
            pct = int(s.win_count / s.task_count * 100)
            win_cells.append(f"{pct}%")
        else:
            win_cells.append("-")
    if has_verify:
        win_cells.append("-")
    lines.append("| " + " | ".join(win_cells) + " |")

    # Avg time
    time_cells = ["**Avg Time**"]
    for agent in agents:
        s = agent_stat_map.get(agent)
        time_cells.append(f"{s.avg_time:.1f}s" if s else "-")
    if has_verify:
        time_cells.append("-")
    lines.append("| " + " | ".join(time_cells) + " |")

    # Total cost
    cost_cells = ["**Total Cost**"]
    for agent in agents:
        s = agent_stat_map.get(agent)
        if s and s.total_cost is not None:
            cost_cells.append(f"${s.total_cost:.4f}")
        else:
            cost_cells.append("-")
    if has_verify:
        cost_cells.append("-")
    lines.append("| " + " | ".join(cost_cells) + " |")

    lines.append("")

    # Task insights
    if stats.task_stats:
        lines.append("## Task Insights\n")
        lines.append("| Task | Best Agent | Best Score | Avg Score | Fastest Agent |")
        lines.append("|------|-----------|-----------|----------|--------------|")
        for ts in stats.task_stats:
            fastest = f"{ts.fastest_agent} ({ts.fastest_time:.0f}s)" if ts.fastest_agent else "-"
            lines.append(
                f"| {ts.task_name} | {ts.best_agent or '-'} | "
                f"{ts.best_score:.1f} | {ts.avg_score:.1f} | {fastest} |"
            )
        lines.append("")

    if has_verify:
        details = _verify_details_rows(result, lookup)
        if details:
            lines.append("## Verification Details\n")
            lines.append("| Task | Agent | Verify | Output |")
            lines.append("|------|-------|--------|--------|")
            for task_name, agent, verify_pct, output in details:
                safe_output = (output or "-").replace("|", "\\|").replace("\n", "<br>")
                lines.append(
                    f"| {task_name} | {agent} | {verify_pct} | {safe_output} |"
                )
            lines.append("")

    elo_section = _render_elo_markdown(elo_ratings or {})
    if elo_section:
        lines.append(elo_section)

    return "\n".join(lines)


def render_benchmark_html(
    result: BenchmarkResult,
    stats: BenchmarkStats,
    elo_ratings: dict[str, float] | None = None,
) -> str:
    """Render a self-contained HTML report for the benchmark results."""
    if _is_multi_trial(result):
        return _render_benchmark_html_trials(result, elo_ratings=elo_ratings)
    agents = result.agents
    tasks = result.tasks
    lookup = {(r.task_name, r.agent): r for r in result.results}
    agent_stat_map = {s.agent: s for s in stats.agent_stats}
    has_verify = _has_verification(result)
    now = datetime.now(timezone.utc).strftime("%Y-%m-%d %H:%M UTC")

    def esc(s: str) -> str:
        return _html_mod.escape(str(s))

    # Build task rows
    task_rows = ""
    for task_name in tasks:
        task_rows += f"<tr><td class='task'>{esc(task_name)}</td>"
        for agent in agents:
            r = lookup.get((task_name, agent))
            if r is None:
                task_rows += "<td>-</td>"
            elif r.error:
                task_rows += "<td class='fail'>ERR</td>"
            elif r.timed_out:
                task_rows += "<td class='warn'>TIMEOUT</td>"
            else:
                cls = "pass" if r.score >= 50 else "fail"
                task_rows += f"<td class='{cls}'>{r.score:.1f}<br><small>({r.wall_time:.0f}s)</small></td>"
        if has_verify:
            verify = _task_verify_percentage(result, task_name, lookup)
            verify_cls = "pass" if verify == "100%" else ("fail" if verify == "0%" else "")
            task_rows += f"<td class='{verify_cls}'>{esc(verify)}</td>"
        task_rows += "</tr>\n"

    # Summary rows
    total_cells = ""
    for agent in agents:
        s = agent_stat_map.get(agent)
        total_cells += f"<td><strong>{s.total_score:.1f}</strong></td>" if s else "<td>-</td>"

    win_cells = ""
    for agent in agents:
        s = agent_stat_map.get(agent)
        if s and s.task_count > 0:
            pct = int(s.win_count / s.task_count * 100)
            win_cells += f"<td>{pct}%</td>"
        else:
            win_cells += "<td>-</td>"

    time_cells = ""
    for agent in agents:
        s = agent_stat_map.get(agent)
        time_cells += f"<td>{s.avg_time:.1f}s</td>" if s else "<td>-</td>"

    cost_cells = ""
    for agent in agents:
        s = agent_stat_map.get(agent)
        if s and s.total_cost is not None:
            cost_cells += f"<td>${s.total_cost:.4f}</td>"
        else:
            cost_cells += "<td>-</td>"

    summary_verify_cell = "<td>-</td>" if has_verify else ""

    agent_headers = "".join(f"<th>{esc(a)}</th>" for a in agents)
    verify_header = "<th>Verify</th>" if has_verify else ""
    winner = stats.agent_stats[0].agent if stats.agent_stats else "-"

    verify_details_section = ""
    if has_verify:
        detail_rows = ""
        for task_name, agent, verify_pct, output in _verify_details_rows(result, lookup):
            css = "pass" if verify_pct == "100%" else ("fail" if verify_pct == "0%" else "")
            formatted_output = esc(output or "-").replace("\n", "<br>")
            detail_rows += (
                f"<tr><td class='task'>{esc(task_name)}</td><td>{esc(agent)}</td>"
                f"<td class='{css}'>{esc(verify_pct)}</td><td><code>{formatted_output}</code></td></tr>\n"
            )
        if detail_rows:
            verify_details_section = f"""
<h2>Verification Details</h2>
<table>
<thead><tr><th>Task</th><th>Agent</th><th>Verify</th><th>Output</th></tr></thead>
<tbody>
{detail_rows}
</tbody>
</table>
"""

    elo_section = _render_elo_html(elo_ratings or {})

    return f"""<!DOCTYPE html>
<html lang="en">
<head>
<meta charset="UTF-8">
<meta name="viewport" content="width=device-width, initial-scale=1.0">
<title>coderace Benchmark — {esc(result.benchmark_id)}</title>
<style>
  body {{ font-family: system-ui, sans-serif; background: #0d1117; color: #c9d1d9; margin: 0; padding: 2rem; }}
  h1, h2 {{ color: #58a6ff; }}
  .meta {{ color: #8b949e; font-size: 0.9rem; margin-bottom: 1.5rem; }}
  .winner {{ background: #1f6feb33; border: 1px solid #1f6feb; border-radius: 6px; padding: 0.75rem 1rem; margin-bottom: 1.5rem; }}
  table {{ border-collapse: collapse; width: 100%; margin-bottom: 2rem; }}
  th {{ background: #161b22; color: #58a6ff; padding: 0.6rem 1rem; border: 1px solid #30363d; }}
  td {{ padding: 0.5rem 1rem; border: 1px solid #30363d; text-align: center; }}
  td.task {{ text-align: left; font-weight: 500; color: #e6edf3; }}
  td.pass {{ color: #3fb950; }}
  td.fail {{ color: #f85149; }}
  td.warn {{ color: #d29922; }}
  tr.summary td {{ background: #161b22; font-weight: bold; color: #e6edf3; }}
  small {{ color: #8b949e; }}
</style>
</head>
<body>
<h1>coderace Benchmark</h1>
<div class="meta">
  ID: {esc(result.benchmark_id)} &nbsp;|&nbsp; Generated: {now}
</div>
<div class="winner">
  <strong>Winner:</strong> {esc(winner)}
</div>
<h2>Results</h2>
<table>
<thead><tr><th>Task</th>{agent_headers}{verify_header}</tr></thead>
<tbody>
{task_rows}
<tr class="summary"><td>TOTAL</td>{total_cells}{summary_verify_cell}</tr>
<tr class="summary"><td>Win Rate</td>{win_cells}{summary_verify_cell}</tr>
<tr class="summary"><td>Avg Time</td>{time_cells}{summary_verify_cell}</tr>
<tr class="summary"><td>Total Cost</td>{cost_cells}{summary_verify_cell}</tr>
</tbody>
</table>
{verify_details_section}
{elo_section}
</body>
</html>
"""



"""Built-in task library for coderace."""

from __future__ import annotations

from importlib import resources
from pathlib import Path

import yaml


def _tasks_path() -> Path:
    """Return the path to the bundled tasks directory."""
    return Path(str(resources.files("coderace.builtins") / "tasks"))


def list_builtins() -> list[str]:
    """List all available built-in task names (without .yaml extension)."""
    tasks_dir = _tasks_path()
    if not tasks_dir.is_dir():
        return []
    return sorted(p.stem for p in tasks_dir.glob("*.yaml"))


def list_builtin_tasks() -> list[str]:
    """Backward-compatible alias for listing built-in task names."""
    return list_builtins()


def load_builtin(name: str) -> dict:
    """Load a built-in task by name and return its parsed YAML as a dict."""
    path = get_builtin_path(name)
    with open(path) as f:
        data = yaml.safe_load(f)
    if not isinstance(data, dict):
        raise ValueError(f"Built-in task {name!r} is not a valid YAML mapping")
    return data


def get_builtin_path(name: str) -> Path:
    """Return the filesystem path to a built-in task YAML file."""
    path = _tasks_path() / f"{name}.yaml"
    if not path.exists():
        available = list_builtins()
        raise FileNotFoundError(
            f"Built-in task {name!r} not found. Available: {', '.join(available) or '(none)'}"
        )
    return path



"""Tests for D3: coderace leaderboard command."""

from __future__ import annotations

import json
from pathlib import Path
from unittest.mock import patch

import pytest
from typer.testing import CliRunner

from coderace.cli import app
from coderace.commands.leaderboard import (
    format_leaderboard_html,
    format_leaderboard_json,
    format_leaderboard_markdown,
    format_leaderboard_terminal,
)
from coderace.store import AgentStat, ResultStore

runner = CliRunner()


def _make_stat(
    agent: str,
    wins: int = 1,
    races: int = 2,
    avg_score: float = 75.0,
    avg_cost: float | None = 0.05,
    avg_time: float = 12.0,
) -> AgentStat:
    return AgentStat(
        agent=agent,
        wins=wins,
        races=races,
        win_rate=wins / races if races > 0 else 0.0,
        avg_score=avg_score,
        avg_cost=avg_cost,
        avg_time=avg_time,
    )


def _populate_store(store: ResultStore) -> None:
    """Add sample data to the store."""
    # Run 1: claude wins
    store.save_run("task-a", [
        {"agent": "claude", "composite_score": 85.0, "wall_time": 10.0,
         "lines_changed": 42, "tests_pass": True, "exit_clean": True,
         "lint_clean": True, "cost_usd": 0.05},
        {"agent": "codex", "composite_score": 70.0, "wall_time": 15.0,
         "lines_changed": 98, "tests_pass": True, "exit_clean": True,
         "lint_clean": False, "cost_usd": 0.03},
    ])
    # Run 2: codex wins
    store.save_run("task-b", [
        {"agent": "claude", "composite_score": 60.0, "wall_time": 12.0,
         "lines_changed": 50, "tests_pass": False, "exit_clean": True,
         "lint_clean": True, "cost_usd": 0.04},
        {"agent": "codex", "composite_score": 80.0, "wall_time": 11.0,
         "lines_changed": 30, "tests_pass": True, "exit_clean": True,
         "lint_clean": True, "cost_usd": 0.06},
    ])
    # Run 3: claude wins again
    store.save_run("task-a", [
        {"agent": "claude", "composite_score": 90.0, "wall_time": 8.0,
         "lines_changed": 35, "tests_pass": True, "exit_clean": True,
         "lint_clean": True, "cost_usd": 0.06},
        {"agent": "codex", "composite_score": 65.0, "wall_time": 14.0,
         "lines_changed": 88, "tests_pass": True, "exit_clean": False,
         "lint_clean": True, "cost_usd": 0.02},
    ])


@pytest.fixture
def populated_store(tmp_path: Path) -> ResultStore:
    db_path = tmp_path / "test.db"
    store = ResultStore(db_path=db_path)
    _populate_store(store)
    yield store
    store.close()


class TestFormatFunctions:
    def test_terminal_empty(self) -> None:
        output = format_leaderboard_terminal([], console=None)
        assert "leaderboard" in output.lower()

    def test_terminal_with_data(self) -> None:
        stats = [_make_stat("claude", wins=2, races=3), _make_stat("codex", wins=1, races=3)]
        output = format_leaderboard_terminal(stats)
        assert "claude" in output
        assert "codex" in output
        assert "67%" in output  # claude win rate

    def test_markdown_empty(self) -> None:
        output = format_leaderboard_markdown([])
        assert "No data" in output

    def test_markdown_with_data(self) -> None:
        stats = [_make_stat("claude", wins=2, races=3)]
        output = format_leaderboard_markdown(stats)
        assert "| 1 |" in output
        assert "`claude`" in output
        assert "67%" in output

    def test_json_format(self) -> None:
        stats = [_make_stat("claude", wins=2, races=3, avg_score=78.3)]
        output = format_leaderboard_json(stats)
        data = json.loads(output)
        assert "leaderboard" in data
        assert len(data["leaderboard"]) == 1
        entry = data["leaderboard"][0]
        assert entry["agent"] == "claude"
        assert entry["wins"] == 2
        assert entry["races"] == 3
        assert entry["avg_score"] == 78.3

    def test_json_null_cost(self) -> None:
        stats = [_make_stat("claude", avg_cost=None)]
        output = format_leaderboard_json(stats)
        data = json.loads(output)
        assert data["leaderboard"][0]["avg_cost"] is None

    def test_html_format(self) -> None:
        stats = [_make_stat("claude")]
        output = format_leaderboard_html(stats)
        assert "<html>" in output
        assert "claude" in output
        assert "leaderboard" in output.lower()

    def test_markdown_cost_none_shows_dash(self) -> None:
        stats = [_make_stat("claude", avg_cost=None)]
        output = format_leaderboard_markdown(stats)
        assert "| - |" in output


class TestLeaderboardCommand:
    def test_empty_db(self, tmp_path: Path) -> None:
        db_path = tmp_path / "empty.db"
        with patch("coderace.store.get_db_path", return_value=db_path):
            result = runner.invoke(app, ["leaderboard"])
        assert result.exit_code == 0
        assert "no data" in result.output.lower()

    def test_default_view(self, populated_store: ResultStore) -> None:
        db_path = populated_store._db_path
        with patch("coderace.store.get_db_path", return_value=db_path):
            result = runner.invoke(app, ["leaderboard"])
        assert result.exit_code == 0
        assert "claude" in result.output
        assert "codex" in result.output

    def test_task_filter(self, populated_store: ResultStore) -> None:
        db_path = populated_store._db_path
        with patch("coderace.store.get_db_path", return_value=db_path):
            result = runner.invoke(app, ["leaderboard", "--task", "task-a"])
        assert result.exit_code == 0
        assert "claude" in result.output

    def test_since_filter(self, populated_store: ResultStore) -> None:
        db_path = populated_store._db_path
        with patch("coderace.store.get_db_path", return_value=db_path):
            # All runs are recent, so 7d should include them
            result = runner.invoke(app, ["leaderboard", "--since", "7d"])
        assert result.exit_code == 0
        assert "claude" in result.output

    def test_min_runs_filter(self, populated_store: ResultStore) -> None:
        db_path = populated_store._db_path
        with patch("coderace.store.get_db_path", return_value=db_path):
            # Both agents have 3 races, require 4 -> empty
            result = runner.invoke(app, ["leaderboard", "--min-runs", "4"])
        assert result.exit_code == 0
        assert "no data" in result.output.lower()

    def test_format_markdown(self, populated_store: ResultStore) -> None:
        db_path = populated_store._db_path
        with patch("coderace.store.get_db_path", return_value=db_path):
            result = runner.invoke(app, ["leaderboard", "--format", "markdown"])
        assert result.exit_code == 0
        assert "| Rank |" in result.output

    def test_format_json(self, populated_store: ResultStore) -> None:
        db_path = populated_store._db_path
        with patch("coderace.store.get_db_path", return_value=db_path):
            result = runner.invoke(app, ["leaderboard", "--format", "json"])
        assert result.exit_code == 0
        data = json.loads(result.output)
        assert "leaderboard" in data

    def test_format_html(self, populated_store: ResultStore) -> None:
        db_path = populated_store._db_path
        with patch("coderace.store.get_db_path", return_value=db_path):
            result = runner.invoke(app, ["leaderboard", "--format", "html"])
        assert result.exit_code == 0
        assert "<html>" in result.output

    def test_invalid_format(self, populated_store: ResultStore) -> None:
        db_path = populated_store._db_path
        with patch("coderace.store.get_db_path", return_value=db_path):
            result = runner.invoke(app, ["leaderboard", "--format", "csv"])
        assert result.exit_code == 1

    def test_ranking_logic(self, populated_store: ResultStore) -> None:
        """Claude has 2 wins/3 races, codex has 1 win/3 races. Claude should rank first."""
        db_path = populated_store._db_path
        with patch("coderace.store.get_db_path", return_value=db_path):
            result = runner.invoke(app, ["leaderboard", "--format", "json"])
        data = json.loads(result.output)
        lb = data["leaderboard"]
        assert lb[0]["agent"] == "claude"
        assert lb[0]["wins"] == 2
        assert lb[1]["agent"] == "codex"
        assert lb[1]["wins"] == 1

    def test_help_text(self) -> None:
        result = runner.invoke(app, ["leaderboard", "--help"])
        assert result.exit_code == 0
        assert "leaderboard" in result.output.lower()
        assert "--task" in result.output
        assert "--since" in result.output
        assert "--min-runs" in result.output
        assert "--format" in result.output



"""SQLite-backed persistent storage for race results."""

from __future__ import annotations

import json
import os
import sqlite3
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import Optional


def _default_db_path() -> Path:
    """Return the default database path (~/.coderace/results.db)."""
    return Path.home() / ".coderace" / "results.db"


def get_db_path() -> Path:
    """Return the database path from env var or default."""
    env = os.environ.get("CODERACE_DB")
    if env:
        return Path(env)
    return _default_db_path()


_SCHEMA = """
CREATE TABLE IF NOT EXISTS runs (
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    task_name TEXT NOT NULL,
    timestamp TEXT NOT NULL,
    git_ref TEXT,
    config_hash TEXT,
    agent_count INTEGER NOT NULL
);

CREATE INDEX IF NOT EXISTS idx_runs_task ON runs(task_name);
CREATE INDEX IF NOT EXISTS idx_runs_timestamp ON runs(timestamp);

CREATE TABLE IF NOT EXISTS agent_results (
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    run_id INTEGER NOT NULL REFERENCES runs(id),
    agent TEXT NOT NULL,
    composite_score REAL NOT NULL,
    wall_time REAL NOT NULL,
    lines_changed INTEGER NOT NULL,
    tests_pass INTEGER NOT NULL,
    exit_clean INTEGER NOT NULL,
    lint_clean INTEGER NOT NULL,
    cost_usd REAL,
    model_name TEXT,
    is_winner INTEGER NOT NULL DEFAULT 0
);

CREATE INDEX IF NOT EXISTS idx_agent_results_run ON agent_results(run_id);
CREATE INDEX IF NOT EXISTS idx_agent_results_agent ON agent_results(agent);

CREATE TABLE IF NOT EXISTS benchmarks (
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    benchmark_id TEXT NOT NULL UNIQUE,
    timestamp TEXT NOT NULL,
    agents TEXT NOT NULL,
    tasks TEXT NOT NULL,
    winner TEXT,
    elapsed REAL
);

CREATE INDEX IF NOT EXISTS idx_benchmarks_timestamp ON benchmarks(timestamp);

CREATE TABLE IF NOT EXISTS benchmark_results (
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    benchmark_id TEXT NOT NULL REFERENCES benchmarks(benchmark_id),
    task_name TEXT NOT NULL,
    agent TEXT NOT NULL,
    trial_number INTEGER NOT NULL DEFAULT 1,
    score REAL NOT NULL,
    wall_time REAL NOT NULL,
    tests_pass INTEGER NOT NULL,
    exit_clean INTEGER NOT NULL,
    lint_clean INTEGER NOT NULL,
    timed_out INTEGER NOT NULL,
    verify_applicable INTEGER NOT NULL DEFAULT 0,
    verify_passed INTEGER NOT NULL DEFAULT 0,
    verify_score REAL NOT NULL DEFAULT 0.0,
    verify_output TEXT NOT NULL DEFAULT '',
    cost_usd REAL,
    error TEXT
);

CREATE INDEX IF NOT EXISTS idx_benchmark_results_bid ON benchmark_results(benchmark_id);

CREATE TABLE IF NOT EXISTS elo_ratings (
    agent TEXT PRIMARY KEY,
    rating REAL NOT NULL DEFAULT 1500,
    updated_at TEXT NOT NULL
);

CREATE INDEX IF NOT EXISTS idx_elo_ratings_rating ON elo_ratings(rating DESC);
"""


@dataclass
class RunRecord:
    """A saved run record."""

    run_id: int
    task_name: str
    timestamp: str
    git_ref: Optional[str]
    agents: list[AgentRecord]


@dataclass
class AgentRecord:
    """A saved agent result record."""

    agent: str
    composite_score: float
    wall_time: float
    lines_changed: int
    tests_pass: bool
    exit_clean: bool
    lint_clean: bool
    cost_usd: Optional[float]
    model_name: Optional[str]
    is_winner: bool


@dataclass
class AgentStat:
    """Aggregate statistics for an agent across runs."""

    agent: str
    wins: int
    races: int
    win_rate: float
    avg_score: float
    avg_cost: Optional[float]
    avg_time: float


class ResultStore:
    """SQLite-backed result store for coderace runs."""

    def __init__(self, db_path: Path | str | None = None) -> None:
        if db_path is None:
            db_path = get_db_path()
        self._db_path = Path(db_path)
        self._db_path.parent.mkdir(parents=True, exist_ok=True)
        self._conn: Optional[sqlite3.Connection] = None
        self._ensure_tables()

    def _get_conn(self) -> sqlite3.Connection:
        if self._conn is None:
            self._conn = sqlite3.connect(str(self._db_path), timeout=10)
            self._conn.execute("PRAGMA journal_mode=WAL")
            self._conn.execute("PRAGMA foreign_keys=ON")
            self._conn.row_factory = sqlite3.Row
        return self._conn

    def _ensure_tables(self) -> None:
        conn = self._get_conn()
        conn.executescript(_SCHEMA)
        self._ensure_benchmark_result_columns(conn)
        conn.commit()

    def _ensure_benchmark_result_columns(self, conn: sqlite3.Connection) -> None:
        """Add benchmark verification columns to older DBs that predate D3."""
        self._ensure_column(
            conn,
            "benchmark_results",
            "verify_applicable",
            "INTEGER NOT NULL DEFAULT 0",
        )
        self._ensure_column(
            conn,
            "benchmark_results",
            "verify_passed",
            "INTEGER NOT NULL DEFAULT 0",
        )
        self._ensure_column(
            conn,
            "benchmark_results",
            "verify_score",
            "REAL NOT NULL DEFAULT 0.0",
        )
        self._ensure_column(
            conn,
            "benchmark_results",
            "verify_output",
            "TEXT NOT NULL DEFAULT ''",
        )
        self._ensure_column(
            conn,
            "benchmark_results",
            "trial_number",
            "INTEGER NOT NULL DEFAULT 1",
        )

    def _ensure_column(
        self,
        conn: sqlite3.Connection,
        table: str,
        column: str,
        column_def: str,
    ) -> None:
        columns = {
            row["name"]
            for row in conn.execute(f"PRAGMA table_info({table})").fetchall()
        }
        if column not in columns:
            conn.execute(f"ALTER TABLE {table} ADD COLUMN {column} {column_def}")

    def close(self) -> None:
        if self._conn is not None:
            self._conn.close()
            self._conn = None

    def save_run(
        self,
        task_name: str,
        results: list[dict],
        git_ref: str | None = None,
        config_hash: str | None = None,
    ) -> int:
        """Save a run with its agent results.

        Args:
            task_name: Name of the task.
            results: List of dicts with keys: agent, composite_score, wall_time,
                     lines_changed, tests_pass, exit_clean, lint_clean,
                     cost_usd (optional), model_name (optional).
            git_ref: Git ref at time of run.
            config_hash: Hash of task configuration.

        Returns:
            The run ID.
        """
        conn = self._get_conn()
        timestamp = datetime.now(timezone.utc).isoformat()

        cursor = conn.execute(
            "INSERT INTO runs (task_name, timestamp, git_ref, config_hash, agent_count) "
            "VALUES (?, ?, ?, ?, ?)",
            (task_name, timestamp, git_ref, config_hash, len(results)),
        )
        run_id = cursor.lastrowid
        assert run_id is not None

        # Determine winner(s): highest composite_score
        if results:
            max_score = max(r["composite_score"] for r in results)
        else:
            max_score = 0.0

        for r in results:
            is_winner = 1 if r["composite_score"] == max_score else 0
            conn.execute(
                "INSERT INTO agent_results "
                "(run_id, agent, composite_score, wall_time, lines_changed, "
                "tests_pass, exit_clean, lint_clean, cost_usd, model_name, is_winner) "
                "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
                (
                    run_id,
                    r["agent"],
                    r["composite_score"],
                    r["wall_time"],
                    r["lines_changed"],
                    1 if r["tests_pass"] else 0,
                    1 if r["exit_clean"] else 0,
                    1 if r["lint_clean"] else 0,
                    r.get("cost_usd"),
                    r.get("model_name"),
                    is_winner,
                ),
            )

        conn.commit()
        return run_id

    def get_runs(
        self,
        task_name: str | None = None,
        agent: str | None = None,
        limit: int = 50,
    ) -> list[RunRecord]:
        """Query past runs.

        Args:
            task_name: Filter by task name.
            agent: Filter to only runs that include this agent.
            limit: Maximum number of runs to return.

        Returns:
            List of RunRecord objects, newest first.
        """
        conn = self._get_conn()

        query = "SELECT DISTINCT r.id, r.task_name, r.timestamp, r.git_ref FROM runs r"
        params: list = []

        if agent:
            query += " JOIN agent_results ar ON ar.run_id = r.id"

        conditions = []
        if task_name:
            conditions.append("r.task_name = ?")
            params.append(task_name)
        if agent:
            conditions.append("ar.agent = ?")
            params.append(agent)

        if conditions:
            query += " WHERE " + " AND ".join(conditions)

        query += " ORDER BY r.timestamp DESC LIMIT ?"
        params.append(limit)

        rows = conn.execute(query, params).fetchall()

        result = []
        for row in rows:
            agents = self._get_agents_for_run(row["id"])
            result.append(
                RunRecord(
                    run_id=row["id"],
                    task_name=row["task_name"],
                    timestamp=row["timestamp"],
                    git_ref=row["git_ref"],
                    agents=agents,
                )
            )
        return result

    def _get_agents_for_run(self, run_id: int) -> list[AgentRecord]:
        conn = self._get_conn()
        rows = conn.execute(
            "SELECT * FROM agent_results WHERE run_id = ? ORDER BY composite_score DESC",
            (run_id,),
        ).fetchall()
        return [
            AgentRecord(
                agent=r["agent"],
                composite_score=r["composite_score"],
                wall_time=r["wall_time"],
                lines_changed=r["lines_changed"],
                tests_pass=bool(r["tests_pass"]),
                exit_clean=bool(r["exit_clean"]),
                lint_clean=bool(r["lint_clean"]),
                cost_usd=r["cost_usd"],
                model_name=r["model_name"],
                is_winner=bool(r["is_winner"]),
            )
            for r in rows
        ]

    def get_agent_stats(
        self,
        agent: str | None = None,
        task_name: str | None = None,
        since: str | None = None,
        min_runs: int = 0,
    ) -> list[AgentStat]:
        """Aggregate statistics per agent.

        Args:
            agent: Filter to a specific agent.
            task_name: Filter to a specific task.
            since: ISO date or relative shorthand ("7d", "30d").
            min_runs: Exclude agents with fewer than this many races.

        Returns:
            List of AgentStat objects, sorted by win rate descending.
        """
        conn = self._get_conn()

        query = (
            "SELECT ar.agent, "
            "COUNT(*) as races, "
            "SUM(ar.is_winner) as wins, "
            "AVG(ar.composite_score) as avg_score, "
            "AVG(ar.cost_usd) as avg_cost, "
            "AVG(ar.wall_time) as avg_time "
            "FROM agent_results ar "
            "JOIN runs r ON r.id = ar.run_id"
        )
        params: list = []
        conditions = []

        if agent:
            conditions.append("ar.agent = ?")
            params.append(agent)
        if task_name:
            conditions.append("r.task_name = ?")
            params.append(task_name)
        if since:
            since_dt = _parse_since(since)
            if since_dt:
                conditions.append("r.timestamp >= ?")
                params.append(since_dt)

        if conditions:
            query += " WHERE " + " AND ".join(conditions)

        query += " GROUP BY ar.agent"

        if min_runs > 0:
            query += " HAVING races >= ?"
            params.append(min_runs)

        query += " ORDER BY wins * 1.0 / COUNT(*) DESC, avg_score DESC"

        rows = conn.execute(query, params).fetchall()
        return [
            AgentStat(
                agent=r["agent"],
                wins=r["wins"],
                races=r["races"],
                win_rate=r["wins"] / r["races"] if r["races"] > 0 else 0.0,
                avg_score=r["avg_score"],
                avg_cost=r["avg_cost"],
                avg_time=r["avg_time"],
            )
            for r in rows
        ]


    def save_benchmark(self, benchmark_result, stats) -> None:
        """Save a BenchmarkResult and its stats to the store."""
        from coderace.benchmark_stats import BenchmarkStats

        conn = self._get_conn()
        ts = datetime.now(timezone.utc).isoformat()
        winner = stats.agent_stats[0].agent if stats.agent_stats else None
        elapsed = benchmark_result.elapsed if benchmark_result.finished_at else None

        agents_str = ",".join(benchmark_result.agents)
        tasks_str = ",".join(benchmark_result.tasks)

        conn.execute(
            "INSERT OR REPLACE INTO benchmarks "
            "(benchmark_id, timestamp, agents, tasks, winner, elapsed) "
            "VALUES (?, ?, ?, ?, ?, ?)",
            (benchmark_result.benchmark_id, ts, agents_str, tasks_str, winner, elapsed),
        )

        for r in benchmark_result.results:
            conn.execute(
                "INSERT INTO benchmark_results "
                "(benchmark_id, task_name, agent, trial_number, score, wall_time, tests_pass, "
                "exit_clean, lint_clean, timed_out, verify_applicable, verify_passed, "
                "verify_score, verify_output, cost_usd, error) "
                "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
                (
                    benchmark_result.benchmark_id,
                    r.task_name,
                    r.agent,
                    r.trial_number,
                    r.score,
                    r.wall_time,
                    1 if r.tests_pass else 0,
                    1 if r.exit_clean else 0,
                    1 if r.lint_clean else 0,
                    1 if r.timed_out else 0,
                    1 if r.verify_applicable else 0,
                    1 if r.verify_passed else 0,
                    r.verify_score,
                    r.verify_output,
                    r.cost_usd,
                    r.error,
                ),
            )
        conn.commit()

    def get_elo_ratings(self) -> dict[str, float]:
        """Return persisted ELO ratings sorted by rating descending."""
        conn = self._get_conn()
        rows = conn.execute(
            "SELECT agent, rating FROM elo_ratings ORDER BY rating DESC, agent ASC"
        ).fetchall()
        return {row["agent"]: float(row["rating"]) for row in rows}

    def upsert_elo_ratings(self, ratings: dict[str, float]) -> None:
        """Insert or update ELO ratings for agents."""
        if not ratings:
            return
        conn = self._get_conn()
        ts = datetime.now(timezone.utc).isoformat()
        for agent, rating in ratings.items():
            conn.execute(
                "INSERT INTO elo_ratings (agent, rating, updated_at) "
                "VALUES (?, ?, ?) "
                "ON CONFLICT(agent) DO UPDATE SET "
                "rating=excluded.rating, updated_at=excluded.updated_at",
                (agent, float(rating), ts),
            )
        conn.commit()

    def reset_elo_ratings(self, initial_rating: float = 1500.0) -> int:
        """Reset all stored ELO ratings to the initial value."""
        conn = self._get_conn()
        ts = datetime.now(timezone.utc).isoformat()
        cursor = conn.execute(
            "UPDATE elo_ratings SET rating = ?, updated_at = ?",
            (float(initial_rating), ts),
        )
        conn.commit()
        return cursor.rowcount

    def get_benchmarks(self, limit: int = 10) -> list[dict]:
        """List past benchmark runs, newest first."""
        conn = self._get_conn()
        rows = conn.execute(
            "SELECT benchmark_id, timestamp, agents, tasks, winner, elapsed "
            "FROM benchmarks ORDER BY timestamp DESC LIMIT ?",
            (limit,),
        ).fetchall()
        result = []
        for row in rows:
            tasks_list = row["tasks"].split(",") if row["tasks"] else []
            result.append({
                "benchmark_id": row["benchmark_id"],
                "timestamp": row["timestamp"],
                "agents": row["agents"],
                "task_count": len(tasks_list),
                "winner": row["winner"] or "-",
                "elapsed": row["elapsed"],
            })
        return result

    def get_benchmark(self, benchmark_id: str) -> dict | None:
        """Retrieve a full benchmark result by ID."""
        conn = self._get_conn()
        row = conn.execute(
            "SELECT * FROM benchmarks WHERE benchmark_id = ?",
            (benchmark_id,),
        ).fetchone()
        if row is None:
            return None

        detail_rows = conn.execute(
            "SELECT * FROM benchmark_results WHERE benchmark_id = ? "
            "ORDER BY task_name ASC, trial_number ASC, agent ASC",
            (benchmark_id,),
        ).fetchall()

        agents = row["agents"].split(",") if row["agents"] else []
        tasks = row["tasks"].split(",") if row["tasks"] else []

        results = [
            {
                "task_name": r["task_name"],
                "agent": r["agent"],
                "trial_number": r["trial_number"],
                "score": r["score"],
                "wall_time": r["wall_time"],
                "tests_pass": bool(r["tests_pass"]),
                "exit_clean": bool(r["exit_clean"]),
                "lint_clean": bool(r["lint_clean"]),
                "timed_out": bool(r["timed_out"]),
                "verify_applicable": bool(r["verify_applicable"]),
                "verify_passed": bool(r["verify_passed"]),
                "verify_score": r["verify_score"],
                "verify_output": r["verify_output"],
                "cost_usd": r["cost_usd"],
                "error": r["error"],
            }
            for r in detail_rows
        ]

        return {
            "benchmark_id": row["benchmark_id"],
            "timestamp": row["timestamp"],
            "agents": agents,
            "tasks": tasks,
            "winner": row["winner"],
            "elapsed": row["elapsed"],
            "results": results,
        }


def _parse_since(since: str) -> str | None:
    """Parse a since string into an ISO datetime string.

    Supports ISO dates (2024-01-01) and relative shorthand (7d, 30d).
    """
    from datetime import timedelta

    since = since.strip()

    # Relative shorthand: "7d", "30d", etc.
    if since.endswith("d") and since[:-1].isdigit():
        days = int(since[:-1])
        dt = datetime.now(timezone.utc) - timedelta(days=days)
        return dt.isoformat()

    # Try ISO date
    try:
        dt = datetime.fromisoformat(since)
        if dt.tzinfo is None:
            dt = dt.replace(tzinfo=timezone.utc)
        return dt.isoformat()
    except ValueError:
        return None



"""Tests for D3: agent:model CLI syntax parsing and adapter factory."""

from __future__ import annotations

import pytest

from coderace.adapters import (
    instantiate_adapter,
    make_display_name,
    parse_agent_spec,
)
from coderace.adapters.claude import ClaudeAdapter
from coderace.adapters.codex import CodexAdapter


# ---------------------------------------------------------------------------
# parse_agent_spec
# ---------------------------------------------------------------------------


def test_parse_agent_spec_no_model() -> None:
    assert parse_agent_spec("codex") == ("codex", None)


def test_parse_agent_spec_with_model() -> None:
    assert parse_agent_spec("codex:gpt-5.4") == ("codex", "gpt-5.4")


def test_parse_agent_spec_claude_with_model() -> None:
    assert parse_agent_spec("claude:opus-4-6") == ("claude", "opus-4-6")


def test_parse_agent_spec_empty_model_returns_none() -> None:
    name, model = parse_agent_spec("codex:")
    assert name == "codex"
    assert model is None


def test_parse_agent_spec_strips_whitespace() -> None:
    name, model = parse_agent_spec(" codex : gpt-5.4 ")
    assert name == "codex"
    assert model == "gpt-5.4"


def test_parse_agent_spec_model_with_colons() -> None:
    # Only first colon splits; model can contain colons
    name, model = parse_agent_spec("codex:provider:model")
    assert name == "codex"
    assert model == "provider:model"


# ---------------------------------------------------------------------------
# make_display_name
# ---------------------------------------------------------------------------


def test_make_display_name_no_model() -> None:
    assert make_display_name("codex", None) == "codex"


def test_make_display_name_with_model() -> None:
    assert make_display_name("codex", "gpt-5.4") == "codex (gpt-5.4)"


def test_make_display_name_empty_model_treated_as_none() -> None:
    # Empty string is falsy
    assert make_display_name("codex", "") == "codex"


# ---------------------------------------------------------------------------
# instantiate_adapter
# ---------------------------------------------------------------------------


def test_instantiate_adapter_no_model() -> None:
    adapter = instantiate_adapter("codex")
    assert isinstance(adapter, CodexAdapter)
    assert adapter.model is None
    assert adapter.name == "codex"


def test_instantiate_adapter_with_model() -> None:
    adapter = instantiate_adapter("codex:gpt-5.4")
    assert isinstance(adapter, CodexAdapter)
    assert adapter.model == "gpt-5.4"
    assert adapter.name == "codex (gpt-5.4)"


def test_instantiate_adapter_claude_with_model() -> None:
    adapter = instantiate_adapter("claude:opus-4-6")
    assert isinstance(adapter, ClaudeAdapter)
    assert adapter.model == "opus-4-6"
    assert adapter.name == "claude (opus-4-6)"


def test_instantiate_adapter_unknown_raises() -> None:
    with pytest.raises(KeyError):
        instantiate_adapter("unknown-agent")


def test_instantiate_adapter_build_command_includes_model() -> None:
    adapter = instantiate_adapter("codex:gpt-5.4")
    cmd = adapter.build_command("do task")
    assert "--model" in cmd
    assert "gpt-5.4" in cmd


def test_instantiate_adapter_no_model_no_flag() -> None:
    adapter = instantiate_adapter("codex")
    cmd = adapter.build_command("do task")
    assert "--model" not in cmd


# ---------------------------------------------------------------------------
# Duplicate agents with different models
# ---------------------------------------------------------------------------


def test_duplicate_agents_different_models() -> None:
    """Same agent type with different models produces distinct display names."""
    a1 = instantiate_adapter("codex:gpt-5.4")
    a2 = instantiate_adapter("codex:gpt-5.3-codex")
    assert a1.name != a2.name
    assert a1.model == "gpt-5.4"
    assert a2.model == "gpt-5.3-codex"


# ---------------------------------------------------------------------------
# Task validation with agent:model syntax
# ---------------------------------------------------------------------------


def test_task_validate_accepts_agent_model_syntax() -> None:
    from pathlib import Path
    from coderace.types import Task

    task = Task(
        name="test",
        description="desc",
        repo=Path("."),
        test_command="pytest",
        agents=["codex:gpt-5.4", "claude:opus-4-6"],
    )
    # Mock repo exists check — validation doesn't check disk
    errors = task.validate()
    # Should have no errors about unknown agents
    agent_errors = [e for e in errors if "Unknown agent" in e]
    assert not agent_errors, f"Got agent errors: {agent_errors}"


def test_task_validate_rejects_unknown_agent_with_model() -> None:
    from pathlib import Path
    from coderace.types import Task

    task = Task(
        name="test",
        description="desc",
        repo=Path("."),
        test_command="pytest",
        agents=["bogus:some-model"],
    )
    errors = task.validate()
    agent_errors = [e for e in errors if "Unknown agent" in e]
    assert agent_errors, "Should have rejected unknown agent"



"""Tests for git operations."""

from __future__ import annotations

from pathlib import Path

from coderace.git_ops import (
    add_worktree,
    branch_name_for,
    checkout,
    create_branch,
    get_current_ref,
    get_diff_stat,
    has_uncommitted_changes,
    remove_worktree,
)


def test_get_current_ref(tmp_repo: Path) -> None:
    ref = get_current_ref(tmp_repo)
    assert len(ref) == 40  # full SHA


def test_create_and_checkout_branch(tmp_repo: Path) -> None:
    base = get_current_ref(tmp_repo)
    create_branch(tmp_repo, "test-branch", base)
    # Should now be on test-branch
    import subprocess

    result = subprocess.run(
        ["git", "branch", "--show-current"], cwd=tmp_repo, capture_output=True, text=True
    )
    assert result.stdout.strip() == "test-branch"


def test_checkout_back(tmp_repo: Path) -> None:
    base = get_current_ref(tmp_repo)
    create_branch(tmp_repo, "feature", base)
    checkout(tmp_repo, "main")
    import subprocess

    result = subprocess.run(
        ["git", "branch", "--show-current"], cwd=tmp_repo, capture_output=True, text=True
    )
    assert result.stdout.strip() == "main"


def test_diff_stat_no_changes(tmp_repo: Path) -> None:
    base = get_current_ref(tmp_repo)
    stat, lines = get_diff_stat(tmp_repo, base)
    assert lines == 0


def test_diff_stat_with_changes(tmp_repo: Path) -> None:
    base = get_current_ref(tmp_repo)
    create_branch(tmp_repo, "changes", base)
    (tmp_repo / "new_file.py").write_text("print('hello')\n")
    import subprocess

    subprocess.run(["git", "add", "."], cwd=tmp_repo, capture_output=True)
    subprocess.run(["git", "commit", "-m", "add file"], cwd=tmp_repo, capture_output=True)
    _, lines = get_diff_stat(tmp_repo, base)
    assert lines > 0


def test_branch_name_for() -> None:
    assert branch_name_for("fix-bug", "claude") == "coderace/claude-fix-bug"
    assert branch_name_for("my-task", "codex") == "coderace/codex-my-task"


def test_has_uncommitted_changes(tmp_repo: Path) -> None:
    assert not has_uncommitted_changes(tmp_repo)
    (tmp_repo / "dirty.txt").write_text("dirty")
    assert has_uncommitted_changes(tmp_repo)


def test_worktree_add_and_remove(tmp_repo: Path, tmp_path: Path) -> None:
    base = get_current_ref(tmp_repo)
    create_branch(tmp_repo, "wt-test", base)
    checkout(tmp_repo, "main")

    wt_path = tmp_path / "worktree"
    add_worktree(tmp_repo, wt_path, "wt-test")
    assert wt_path.exists()
    assert (wt_path / "README.md").exists()

    remove_worktree(tmp_repo, wt_path)



"""Tests for D4: coderace history command."""

from __future__ import annotations

import json
from pathlib import Path
from unittest.mock import patch

import pytest
from typer.testing import CliRunner

from coderace.cli import app
from coderace.commands.history import (
    format_history_json,
    format_history_markdown,
    format_history_terminal,
)
from coderace.store import ResultStore

runner = CliRunner()


def _populate_store(store: ResultStore) -> None:
    """Add sample runs."""
    store.save_run("task-a", [
        {"agent": "claude", "composite_score": 85.0, "wall_time": 10.0,
         "lines_changed": 42, "tests_pass": True, "exit_clean": True,
         "lint_clean": True},
        {"agent": "codex", "composite_score": 70.0, "wall_time": 15.0,
         "lines_changed": 98, "tests_pass": True, "exit_clean": True,
         "lint_clean": False},
    ])
    store.save_run("task-b", [
        {"agent": "aider", "composite_score": 60.0, "wall_time": 12.0,
         "lines_changed": 50, "tests_pass": False, "exit_clean": True,
         "lint_clean": True},
    ])
    store.save_run("task-a", [
        {"agent": "claude", "composite_score": 90.0, "wall_time": 8.0,
         "lines_changed": 35, "tests_pass": True, "exit_clean": True,
         "lint_clean": True},
    ])


@pytest.fixture
def populated_store(tmp_path: Path) -> ResultStore:
    db_path = tmp_path / "test.db"
    store = ResultStore(db_path=db_path)
    _populate_store(store)
    yield store
    store.close()


class TestFormatFunctions:
    def test_terminal_empty(self) -> None:
        output = format_history_terminal([])
        assert "history" in output.lower()

    def test_terminal_with_data(self, populated_store: ResultStore) -> None:
        runs = populated_store.get_runs()
        output = format_history_terminal(runs)
        assert "claude" in output
        assert "task-a" in output

    def test_markdown_empty(self) -> None:
        output = format_history_markdown([])
        assert "No runs" in output

    def test_markdown_with_data(self, populated_store: ResultStore) -> None:
        runs = populated_store.get_runs()
        output = format_history_markdown(runs)
        assert "| Run ID |" in output
        assert "`task-a`" in output

    def test_json_format(self, populated_store: ResultStore) -> None:
        runs = populated_store.get_runs()
        output = format_history_json(runs)
        data = json.loads(output)
        assert "history" in data
        assert len(data["history"]) == 3


class TestHistoryCommand:
    def test_empty_db(self, tmp_path: Path) -> None:
        db_path = tmp_path / "empty.db"
        with patch("coderace.store.get_db_path", return_value=db_path):
            result = runner.invoke(app, ["history"])
        assert result.exit_code == 0
        assert "no runs" in result.output.lower()

    def test_default_view(self, populated_store: ResultStore) -> None:
        db_path = populated_store._db_path
        with patch("coderace.store.get_db_path", return_value=db_path):
            result = runner.invoke(app, ["history"])
        assert result.exit_code == 0
        assert "task-a" in result.output

    def test_task_filter(self, populated_store: ResultStore) -> None:
        db_path = populated_store._db_path
        with patch("coderace.store.get_db_path", return_value=db_path):
            result = runner.invoke(app, ["history", "--task", "task-b"])
        assert result.exit_code == 0
        assert "task-b" in result.output

    def test_agent_filter(self, populated_store: ResultStore) -> None:
        db_path = populated_store._db_path
        with patch("coderace.store.get_db_path", return_value=db_path):
            result = runner.invoke(app, ["history", "--agent", "aider"])
        assert result.exit_code == 0
        assert "task-b" in result.output

    def test_limit(self, populated_store: ResultStore) -> None:
        db_path = populated_store._db_path
        with patch("coderace.store.get_db_path", return_value=db_path):
            result = runner.invoke(app, ["history", "--limit", "1", "--format", "json"])
        data = json.loads(result.output)
        assert len(data["history"]) == 1

    def test_format_markdown(self, populated_store: ResultStore) -> None:
        db_path = populated_store._db_path
        with patch("coderace.store.get_db_path", return_value=db_path):
            result = runner.invoke(app, ["history", "--format", "markdown"])
        assert result.exit_code == 0
        assert "| Run ID |" in result.output

    def test_format_json(self, populated_store: ResultStore) -> None:
        db_path = populated_store._db_path
        with patch("coderace.store.get_db_path", return_value=db_path):
            result = runner.invoke(app, ["history", "--format", "json"])
        assert result.exit_code == 0
        data = json.loads(result.output)
        assert "history" in data

    def test_invalid_format(self, populated_store: ResultStore) -> None:
        db_path = populated_store._db_path
        with patch("coderace.store.get_db_path", return_value=db_path):
            result = runner.invoke(app, ["history", "--format", "csv"])
        assert result.exit_code == 1

    def test_help_text(self) -> None:
        result = runner.invoke(app, ["history", "--help"])
        assert result.exit_code == 0
        assert "--task" in result.output
        assert "--agent" in result.output
        assert "--limit" in result.output
        assert "--format" in result.output

    def test_json_includes_winner(self, populated_store: ResultStore) -> None:
        db_path = populated_store._db_path
        with patch("coderace.store.get_db_path", return_value=db_path):
            result = runner.invoke(app, ["history", "--format", "json"])
        data = json.loads(result.output)
        # First run (newest) has claude as sole agent and winner
        run = data["history"][0]
        assert run["winner"] == "claude"

    def test_newest_first(self, populated_store: ResultStore) -> None:
        db_path = populated_store._db_path
        with patch("coderace.store.get_db_path", return_value=db_path):
            result = runner.invoke(app, ["history", "--format", "json"])
        data = json.loads(result.output)
        # Run IDs should be descending (newest first)
        ids = [r["run_id"] for r in data["history"]]
        assert ids == sorted(ids, reverse=True)



"""Tests for task loading and validation."""

from __future__ import annotations

from pathlib import Path

import pytest

from coderace.task import create_template, load_task
from coderace.types import Task, VERIFY_AWARE_DEFAULT_WEIGHTS


def test_load_valid_task(task_yaml: Path) -> None:
    task = load_task(task_yaml)
    assert task.name == "test-task"
    assert task.description == "Fix the bug in main.py"
    assert task.test_command == 'echo "tests pass"'
    assert task.agents == ["claude", "codex"]
    assert task.timeout == 60


def test_load_missing_file() -> None:
    with pytest.raises(FileNotFoundError):
        load_task("/nonexistent/path.yaml")


def test_load_missing_required_fields(tmp_path: Path) -> None:
    path = tmp_path / "bad.yaml"
    path.write_text("name: test\n")
    with pytest.raises(ValueError, match="Missing required fields"):
        load_task(path)


def test_load_invalid_yaml(tmp_path: Path) -> None:
    path = tmp_path / "bad.yaml"
    path.write_text("just a string\n")
    with pytest.raises(ValueError, match="YAML mapping"):
        load_task(path)


def test_load_unknown_agent(tmp_path: Path) -> None:
    path = tmp_path / "bad.yaml"
    path.write_text("""name: t
description: d
test_command: echo ok
agents:
  - unknown_agent
""")
    with pytest.raises(ValueError, match="Unknown agent"):
        load_task(path)


def test_load_default_timeout(tmp_path: Path) -> None:
    path = tmp_path / "task.yaml"
    path.write_text("""name: t
description: d
test_command: echo ok
agents:
  - claude
""")
    task = load_task(path)
    assert task.timeout == 300


def test_create_template(tmp_path: Path) -> None:
    path = create_template("my-task", tmp_path)
    assert path.exists()
    assert path.name == "my-task.yaml"
    content = path.read_text()
    assert "my-task" in content
    assert "agents:" in content


def test_task_validate_empty_name() -> None:
    task = Task(name="", description="d", repo=Path("."), test_command="echo", agents=["claude"])
    errors = task.validate()
    assert any("name" in e for e in errors)


def test_task_validate_no_agents() -> None:
    task = Task(name="t", description="d", repo=Path("."), test_command="echo", agents=[])
    errors = task.validate()
    assert any("agent" in e.lower() for e in errors)


def test_load_task_with_verification_fields(tmp_path: Path) -> None:
    path = tmp_path / "task.yaml"
    path.write_text("""name: t
description: d
test_command: echo ok
verify_command: python3 -m pytest verify_tests.py -x -q
verify_files:
  verify_tests.py: |
    def test_ok():
      assert True
agents:
  - claude
""")
    task = load_task(path)
    assert task.verify_command == "python3 -m pytest verify_tests.py -x -q"
    assert task.verify_files == {
        "verify_tests.py": "def test_ok():\n  assert True\n"
    }
    assert task.get_weights() == VERIFY_AWARE_DEFAULT_WEIGHTS


def test_load_task_rejects_non_mapping_verify_files(tmp_path: Path) -> None:
    path = tmp_path / "task.yaml"
    path.write_text("""name: t
description: d
test_command: echo ok
verify_command: echo ok
verify_files:
  - bad
agents:
  - claude
""")
    with pytest.raises(ValueError, match="verify_files must be a mapping"):
        load_task(path)



"""Base adapter interface."""

from __future__ import annotations

import subprocess
import time
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional

from coderace.cost import CostResult
from coderace.types import AgentResult


class BaseAdapter(ABC):
    """Abstract base class for coding agent adapters."""

    name: str = "base"

    def __init__(self, model: Optional[str] = None) -> None:
        """Initialize adapter with optional model override."""
        self.model = model

    @abstractmethod
    def build_command(self, task_description: str, model: Optional[str] = None) -> list[str]:
        """Build the CLI command to invoke this agent."""
        ...

    def parse_cost(
        self,
        stdout: str,
        stderr: str,
        model_name: str = "",
        custom_pricing: dict[str, tuple[float, float]] | None = None,
    ) -> Optional[CostResult]:
        """Parse cost data from agent output. Override in subclasses.

        Returns None if cost data is unavailable.
        """
        return None

    def run(
        self,
        task_description: str,
        workdir: Path,
        timeout: int,
        no_cost: bool = False,
        custom_pricing: dict[str, tuple[float, float]] | None = None,
    ) -> AgentResult:
        """Run the agent on a task and capture results."""
        model = self.model
        cmd = self.build_command(task_description, model=model)
        start = time.monotonic()
        timed_out = False

        try:
            proc = subprocess.run(
                cmd,
                cwd=workdir,
                capture_output=True,
                text=True,
                timeout=timeout,
            )
            exit_code = proc.returncode
            stdout = proc.stdout
            stderr = proc.stderr
        except subprocess.TimeoutExpired as e:
            timed_out = True
            exit_code = -1
            stdout = (e.stdout or b"").decode("utf-8", errors="replace") if e.stdout else ""
            stderr = (e.stderr or b"").decode("utf-8", errors="replace") if e.stderr else ""
        except FileNotFoundError:
            timed_out = False
            exit_code = 127
            stdout = ""
            stderr = f"Command not found: {cmd[0]}"

        wall_time = time.monotonic() - start

        # Parse cost (fails gracefully — never raises)
        cost_result: Optional[CostResult] = None
        if not no_cost:
            try:
                cost_result = self.parse_cost(
                    stdout,
                    stderr,
                    model_name=model or "",
                    custom_pricing=custom_pricing,
                )
            except Exception:
                pass

        return AgentResult(
            agent=self.name,
            exit_code=exit_code,
            stdout=stdout,
            stderr=stderr,
            wall_time=wall_time,
            timed_out=timed_out,
            cost_result=cost_result,
        )



"""Tests for built-in task library."""

from __future__ import annotations

import pytest

from coderace.builtins import get_builtin_path, list_builtins, load_builtin


EXPECTED_BUILTINS = [
    "api-client",
    "binary-search-tree",
    "bug-hunt",
    "cli-args-parser",
    "concurrent-queue",
    "csv-analyzer",
    "data-pipeline",
    "diff-algorithm",
    "expression-evaluator",
    "fibonacci",
    "file-watcher",
    "http-server",
    "json-parser",
    "lru-cache",
    "markdown-to-html",
    "refactor",
    "regex-engine",
    "state-machine",
    "task-scheduler",
    "url-router",
]


def test_list_builtins_returns_list() -> None:
    names = list_builtins()
    assert isinstance(names, list)
    assert len(names) >= 20
    for expected in EXPECTED_BUILTINS:
        assert expected in names


def test_list_builtins_sorted() -> None:
    names = list_builtins()
    assert names == sorted(names)


def test_load_builtin_fibonacci() -> None:
    data = load_builtin("fibonacci")
    assert isinstance(data, dict)
    assert data["name"] == "fibonacci"
    assert "description" in data
    assert "test_command" in data
    assert "agents" in data


def test_load_builtin_has_required_fields() -> None:
    """Every built-in task must have the required fields."""
    required = {"name", "description", "test_command", "agents"}
    for name in list_builtins():
        data = load_builtin(name)
        missing = required - set(data.keys())
        assert not missing, f"Built-in {name!r} missing fields: {missing}"


def test_load_builtin_missing_task() -> None:
    with pytest.raises(FileNotFoundError, match="not found"):
        load_builtin("nonexistent-task-xyz")


def test_get_builtin_path_exists() -> None:
    path = get_builtin_path("fibonacci")
    assert path.exists()
    assert path.suffix == ".yaml"


def test_get_builtin_path_missing() -> None:
    with pytest.raises(FileNotFoundError, match="not found"):
        get_builtin_path("nonexistent-task-xyz")


def test_all_builtins_valid_yaml() -> None:
    """Every built-in task must be valid YAML with all expected fields."""
    required = {"name", "description", "test_command", "agents"}
    for name in EXPECTED_BUILTINS:
        data = load_builtin(name)
        missing = required - set(data.keys())
        assert not missing, f"Built-in {name!r} missing: {missing}"
        assert isinstance(data["agents"], list)
        assert len(data["agents"]) >= 1
        assert isinstance(data.get("timeout", 300), int)
        assert isinstance(data.get("scoring", {}), dict)



"""Tests for batch 2 benchmark tasks: bug-hunt, refactor, concurrent-queue, api-client."""

from __future__ import annotations

import pytest

from coderace.builtins import get_builtin_path, load_builtin
from coderace.task import load_task


NEW_TASKS = ["bug-hunt", "refactor", "concurrent-queue", "api-client"]


@pytest.mark.parametrize("task_name", NEW_TASKS)
def test_task_loads_as_builtin(task_name):
    data = load_builtin(task_name)
    assert data["name"] == task_name
    assert "description" in data
    assert "test_command" in data
    assert "agents" in data


@pytest.mark.parametrize("task_name", NEW_TASKS)
def test_task_has_verify_suite(task_name):
    data = load_builtin(task_name)
    assert "verify_command" in data
    assert "verify_files" in data
    assert isinstance(data["verify_files"], dict)
    assert len(data["verify_files"]) >= 1


@pytest.mark.parametrize("task_name", NEW_TASKS)
def test_task_loads_via_load_task(task_name):
    path = get_builtin_path(task_name)
    task = load_task(path)
    assert task.name == task_name
    assert task.verify_command is not None
    assert task.verify_files is not None
    assert task.timeout > 0
    assert len(task.agents) >= 1


@pytest.mark.parametrize("task_name", NEW_TASKS)
def test_task_has_scoring_weights(task_name):
    data = load_builtin(task_name)
    assert "scoring" in data
    scoring = data["scoring"]
    assert isinstance(scoring, dict)
    assert "tests" in scoring or "tests_pass" in scoring
    assert "verify" in scoring or "verify_passed" in scoring


@pytest.mark.parametrize("task_name", NEW_TASKS)
def test_task_has_lint_command(task_name):
    data = load_builtin(task_name)
    assert "lint_command" in data
    assert "ruff" in data["lint_command"]


@pytest.mark.parametrize("task_name", NEW_TASKS)
def test_task_yaml_valid(task_name):
    path = get_builtin_path(task_name)
    task = load_task(path)
    errors = task.validate()
    assert not errors, f"Validation errors for {task_name}: {errors}"


@pytest.mark.parametrize("task_name", NEW_TASKS)
def test_task_difficulty_is_hard(task_name):
    data = load_builtin(task_name)
    assert data.get("difficulty") == "hard"


def test_all_new_tasks_in_builtins_list():
    from coderace.builtins import list_builtins
    builtins = list_builtins()
    for task_name in NEW_TASKS:
        assert task_name in builtins


def test_total_builtins_at_least_20():
    from coderace.builtins import list_builtins
    assert len(list_builtins()) >= 20



"""Tests for coderace.store — ResultStore."""

from __future__ import annotations

import os
import sqlite3
import threading
from pathlib import Path

import pytest

from coderace.store import ResultStore, get_db_path


def _make_result(agent: str, score: float = 75.0, **kwargs) -> dict:
    """Helper to build a minimal agent result dict."""
    base = {
        "agent": agent,
        "composite_score": score,
        "wall_time": 10.5,
        "lines_changed": 42,
        "tests_pass": True,
        "exit_clean": True,
        "lint_clean": True,
    }
    base.update(kwargs)
    return base


@pytest.fixture
def store(tmp_path: Path) -> ResultStore:
    db_path = tmp_path / "test.db"
    s = ResultStore(db_path=db_path)
    yield s
    s.close()


class TestSaveRun:
    def test_save_and_returns_id(self, store: ResultStore) -> None:
        run_id = store.save_run("task-1", [_make_result("claude")])
        assert isinstance(run_id, int)
        assert run_id >= 1

    def test_save_multiple_agents(self, store: ResultStore) -> None:
        results = [
            _make_result("claude", score=85.0),
            _make_result("codex", score=70.0),
        ]
        run_id = store.save_run("task-1", results)
        runs = store.get_runs()
        assert len(runs) == 1
        assert len(runs[0].agents) == 2

    def test_save_empty_results(self, store: ResultStore) -> None:
        run_id = store.save_run("task-1", [])
        runs = store.get_runs()
        assert len(runs) == 1
        assert len(runs[0].agents) == 0

    def test_winner_detection(self, store: ResultStore) -> None:
        results = [
            _make_result("claude", score=85.0),
            _make_result("codex", score=70.0),
            _make_result("aider", score=60.0),
        ]
        store.save_run("task-1", results)
        runs = store.get_runs()
        agents = runs[0].agents
        winners = [a for a in agents if a.is_winner]
        assert len(winners) == 1
        assert winners[0].agent == "claude"

    def test_tied_winners(self, store: ResultStore) -> None:
        results = [
            _make_result("claude", score=80.0),
            _make_result("codex", score=80.0),
        ]
        store.save_run("task-1", results)
        runs = store.get_runs()
        winners = [a for a in runs[0].agents if a.is_winner]
        assert len(winners) == 2


class TestGetRuns:
    def test_empty_db(self, store: ResultStore) -> None:
        runs = store.get_runs()
        assert runs == []

    def test_filter_by_task(self, store: ResultStore) -> None:
        store.save_run("task-a", [_make_result("claude")])
        store.save_run("task-b", [_make_result("codex")])

        runs = store.get_runs(task_name="task-a")
        assert len(runs) == 1
        assert runs[0].task_name == "task-a"

    def test_filter_by_agent(self, store: ResultStore) -> None:
        store.save_run("task-1", [_make_result("claude"), _make_result("codex")])
        store.save_run("task-2", [_make_result("aider")])

        runs = store.get_runs(agent="claude")
        assert len(runs) == 1
        assert runs[0].task_name == "task-1"

    def test_limit(self, store: ResultStore) -> None:
        for i in range(10):
            store.save_run(f"task-{i}", [_make_result("claude")])

        runs = store.get_runs(limit=3)
        assert len(runs) == 3

    def test_newest_first(self, store: ResultStore) -> None:
        store.save_run("task-old", [_make_result("claude")])
        store.save_run("task-new", [_make_result("claude")])

        runs = store.get_runs()
        # Newest first
        assert runs[0].task_name == "task-new"
        assert runs[1].task_name == "task-old"

    def test_git_ref_saved(self, store: ResultStore) -> None:
        store.save_run("task-1", [_make_result("claude")], git_ref="abc123")
        runs = store.get_runs()
        assert runs[0].git_ref == "abc123"


class TestGetAgentStats:
    def test_empty_db(self, store: ResultStore) -> None:
        stats = store.get_agent_stats()
        assert stats == []

    def test_single_run(self, store: ResultStore) -> None:
        store.save_run("task-1", [
            _make_result("claude", score=85.0),
            _make_result("codex", score=70.0),
        ])
        stats = store.get_agent_stats()
        assert len(stats) == 2
        claude_stat = next(s for s in stats if s.agent == "claude")
        assert claude_stat.wins == 1
        assert claude_stat.races == 1
        assert claude_stat.win_rate == 1.0
        assert claude_stat.avg_score == 85.0

    def test_multi_run_stats(self, store: ResultStore) -> None:
        # Claude wins first, codex wins second
        store.save_run("task-1", [
            _make_result("claude", score=90.0),
            _make_result("codex", score=70.0),
        ])
        store.save_run("task-2", [
            _make_result("claude", score=60.0),
            _make_result("codex", score=80.0),
        ])
        stats = store.get_agent_stats()
        for s in stats:
            assert s.wins == 1
            assert s.races == 2
            assert s.win_rate == 0.5

    def test_filter_by_agent(self, store: ResultStore) -> None:
        store.save_run("t", [
            _make_result("claude", score=80.0),
            _make_result("codex", score=70.0),
        ])
        stats = store.get_agent_stats(agent="claude")
        assert len(stats) == 1
        assert stats[0].agent == "claude"

    def test_filter_by_task(self, store: ResultStore) -> None:
        store.save_run("task-a", [_make_result("claude", score=80.0)])
        store.save_run("task-b", [_make_result("claude", score=60.0)])
        stats = store.get_agent_stats(task_name="task-a")
        assert len(stats) == 1
        assert stats[0].avg_score == 80.0

    def test_min_runs_filter(self, store: ResultStore) -> None:
        store.save_run("t", [_make_result("claude"), _make_result("codex")])
        store.save_run("t", [_make_result("claude")])
        # claude has 2 races, codex has 1
        stats = store.get_agent_stats(min_runs=2)
        assert len(stats) == 1
        assert stats[0].agent == "claude"

    def test_cost_tracking(self, store: ResultStore) -> None:
        store.save_run("t", [
            _make_result("claude", cost_usd=0.05, model_name="claude-sonnet-4-6"),
        ])
        stats = store.get_agent_stats()
        assert stats[0].avg_cost == pytest.approx(0.05)

    def test_since_filter(self, store: ResultStore) -> None:
        # Insert a run, then filter with a future date (should return 0)
        store.save_run("t", [_make_result("claude")])
        stats = store.get_agent_stats(since="2099-01-01")
        assert stats == []


class TestEdgeCases:
    def test_auto_create_db(self, tmp_path: Path) -> None:
        db_path = tmp_path / "sub" / "dir" / "test.db"
        store = ResultStore(db_path=db_path)
        assert db_path.exists()
        store.close()

    def test_env_var_db_path(self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
        db = tmp_path / "env.db"
        monkeypatch.setenv("CODERACE_DB", str(db))
        assert get_db_path() == db

    def test_default_db_path(self, monkeypatch: pytest.MonkeyPatch) -> None:
        monkeypatch.delenv("CODERACE_DB", raising=False)
        path = get_db_path()
        assert path == Path.home() / ".coderace" / "results.db"

    def test_concurrent_writes(self, tmp_path: Path) -> None:
        db_path = tmp_path / "concurrent.db"
        errors: list[Exception] = []

        def write_run(idx: int) -> None:
            try:
                s = ResultStore(db_path=db_path)
                s.save_run(f"task-{idx}", [_make_result("claude", score=float(idx))])
                s.close()
            except Exception as e:
                errors.append(e)

        threads = [threading.Thread(target=write_run, args=(i,)) for i in range(5)]
        for t in threads:
            t.start()
        for t in threads:
            t.join()

        assert errors == [], f"Concurrent write errors: {errors}"

        # Verify all runs saved
        s = ResultStore(db_path=db_path)
        runs = s.get_runs(limit=100)
        assert len(runs) == 5
        s.close()

    def test_close_and_reopen(self, tmp_path: Path) -> None:
        db_path = tmp_path / "reopen.db"
        store = ResultStore(db_path=db_path)
        store.save_run("t", [_make_result("claude")])
        store.close()

        store2 = ResultStore(db_path=db_path)
        runs = store2.get_runs()
        assert len(runs) == 1
        store2.close()

    def test_agent_fields_round_trip(self, store: ResultStore) -> None:
        result = _make_result(
            "claude",
            score=85.5,
            wall_time=12.3,
            lines_changed=99,
            tests_pass=False,
            exit_clean=True,
            lint_clean=False,
            cost_usd=0.123,
            model_name="claude-sonnet-4-6",
        )
        store.save_run("t", [result])
        runs = store.get_runs()
        a = runs[0].agents[0]
        assert a.agent == "claude"
        assert a.composite_score == pytest.approx(85.5)
        assert a.wall_time == pytest.approx(12.3)
        assert a.lines_changed == 99
        assert a.tests_pass is False
        assert a.exit_clean is True
        assert a.lint_clean is False
        assert a.cost_usd == pytest.approx(0.123)
        assert a.model_name == "claude-sonnet-4-6"
